import abc
import collections


class Rotatable(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def rot(self):
        '''integer: stores the rotation in radians'''

    @property
    @abc.abstractmethod
    def flipped(self):
        '''boolean: stores the flipped state'''

    @abc.abstractmethod
    def rotate(self):
        '''rotate the rotatable object'''

    def flip(self):
        '''flip the rotatable object'''

    def update(self):
        '''update the respective GUI element'''


class Image(metaclass=abc.ABCMeta):
    '''create a matplotlib image'''

    @abc.abstractmethod
    def __init__(self, image, figsize: tuple, cmap='gray'):
        '''Arguments:
        image: numpy ndarray
        figsize: tuple (width, heigth) image size in inches
        '''

    @property
    @abc.abstractmethod
    def fig(self):
        '''matplotlib figure'''

    @property
    @abc.abstractmethod
    def ax(self):
        '''matplotlib axes'''

    @abc.abstractmethod
    def update(self, image):
        '''update the image with new data
        Arguments:
        image: numpy ndarray
        '''


class ZoomImage(Image):
    '''create a matplotlib image that can easily be zoomed'''

    @abc.abstractmethod
    def __init__(self, image, zoom=1, cmap='gray'):
        '''Arguments:
        image: numpy ndarray
        zoom: the zoom factor (optional)
        '''


class SaveLoad(metaclass=abc.ABCMeta):
    '''Base class for classes that can save and load data to/from files.'''

    @abc.abstractmethod
    def load(self, path):
        '''load data from file'''

    @abc.abstractmethod
    def save(self, path):
        '''save data to file'''


class IsEmpty(metaclass=abc.ABCMeta):
    '''Interface for classes that can be empty'''
    @abc.abstractmethod
    def empty(self) -> bool:
        '''return True if the class does not contain data, False otherwise'''


class Config(collections.UserDict, SaveLoad, metaclass=abc.ABCMeta):
    '''A Configuration must be an implementation of MutableSequence that accepts only ConfigItems or Config as elements'''

    @property
    def panel_name(self):
        '''string: Name of the configuration panel'''
        return self._panel_name

    @panel_name.setter
    def panel_name(self, name):
        assert isinstance(name, str) or name is None, 'panel_name must be str or None'
        self._panel_name = name

    @abc.abstractmethod
    def new_item(self, name, value, **kw):
        '''Create a new ConfigItem and add it to the Config sequence'''

    @abc.abstractmethod
    def update(self, d):
        '''parse a dict, create and add the respective ConfigItems'''

    def parse(self, val):
        if isinstance(val, (ConfigItem, Config)):
            return val
        else:
            raise ValueError('Config items must be Config or ConfigItem, got {}'.format(type(val)))

    def __setitem__(self, key, val):
        self.data[key] = self.parse(val)


class ConfigItem(metaclass=abc.ABCMeta):
    GUI_ELEMENT_TYPES = ['entry']

    @property
    def name(self):
        '''string: Name of the configuration item'''
        return self._name

    @name.setter
    def name(self, name):
        assert isinstance(name, str)
        self._name = name

    @property
    @abc.abstractmethod
    def value(self):
        '''Any: Value of the configuration item'''

    @property
    @abc.abstractmethod
    def default_value(self):
        '''Any: Default value of the configuration item'''

    @property
    @abc.abstractmethod
    def gui_element_type(self):
        '''The gui element that should be used by the user to set the config item'''

    @property
    @abc.abstractmethod
    def on_change(self):
        '''Callable: A callback that is executed when the value changes'''

    @abc.abstractmethod
    def convert_val(self, val):
        '''Ensure that val is valid for value and default_value.'''

    @abc.abstractmethod
    def is_valid(self, val):
        '''Check that val can be converted into a valid value by convert_val.'''

    @abc.abstractmethod
    def reset(self):
        '''Reset the value to default_value.'''

    @abc.abstractmethod
    def name2var(self) -> str:
        '''convert the name to a variable name'''


class NumericConfigItem(ConfigItem):
    GUI_ELEMENT_TYPES = ['slider', 'spinbox']
    @property
    @abc.abstractmethod
    def increment(self):
        '''increment'''

    @property
    @abc.abstractmethod
    def from_(self):
        '''min value'''

    @property
    @abc.abstractmethod
    def to(self):
        '''max value'''

    @abc.abstractmethod
    def check(self, val):
        '''Ensure that val is valid for from_ and to.'''


class StringConfigItem(ConfigItem):
    pass


class OptionConfigItem(ConfigItem):

    @property
    @abc.abstractmethod
    def options(self):
        '''list of options'''

    @property
    @abc.abstractmethod
    def current_option(self):
        '''the element in options that has the index value.

        the corresponding setter changes value to the index corresponding to its given argument'''

    @abc.abstractmethod
    def toggle(self):
        '''toggle through the options'''

    @property
    @abc.abstractmethod
    def from_(self):
        '''min value (always 0)'''

    @property
    @abc.abstractmethod
    def to(self):
        '''max value (len(options) - 1)'''

    @property
    @abc.abstractmethod
    def on_options_change(self):
        '''callback triggered when options change'''


class Stack(SaveLoad, IsEmpty):
    '''A data stack'''

    @property
    @abc.abstractmethod
    def num_slices(self):
        '''return: int: number of images in the stack'''


class ImageStack(Stack):

    @property
    @abc.abstractmethod
    def width(self):
        '''return: int: width of each image in pixels'''

    @property
    @abc.abstractmethod
    def height(self):
        '''return: int: height of each image in pixels'''

    @property
    @abc.abstractmethod
    def image_shape(self):
        '''return: tuple: shape of each image in pixels'''

    @property
    @abc.abstractmethod
    def num_channels(self):
        '''return: int: number of color channels of each image'''

    @abc.abstractmethod
    def median(self):
        '''return: 2D np.ndarray: median projection of the stack'''

    @abc.abstractmethod
    def mean(self):
        '''return: 2D np.ndarray: mean projection of the stack'''

    @abc.abstractmethod
    def std(self):
        '''return: 2D np.ndarray: standard deviation projection of the stack'''

    @abc.abstractmethod
    def max(self):
        '''return: 2D np.ndarray: maximum projection of the stack'''

    @abc.abstractmethod
    def copy_tile(self, pos, dim):
        '''copy a tile out of the stack.
        Arguments:
        tuple pos: x and y coordinates
        typle dim: wdth and height of the tile
        return: 4D np.ndarray: copy a part of the stack'''


class MultiImageStack(SaveLoad, IsEmpty):
    '''many image stacks stored as collections.MutableSequence'''

    @abc.abstractmethod
    def __len__(self):
        '''return len(self.data)'''

    @abc.abstractmethod
    def __getitem__(self, i):
        '''return self.data[i]'''

    @abc.abstractmethod
    def __delitem__(self, i):
        '''del self.data[i] and self.positions[i]'''

    @abc.abstractmethod
    def __iter__(self):
        '''iterate over self.data'''

    @abc.abstractmethod
    def append(self, stack, position):
        '''append stack to self.data and position to self.stack_positions[i]'''

    @abc.abstractmethod
    def np_array(self):
        '''return np.array(self.data)'''

    @abc.abstractmethod
    def image_stacks(self, key=None):
        '''return a generator that yields ImageStacks. 
        if key is provided, an individual ImageStack is created from the data corresponding to that key.
        '''


class ImageStackClassification(Stack):
    '''Image stack classification'''

    @property
    @abc.abstractmethod
    def category_names(self):
        '''generator that returns category names.
        names are either generated on the fly or returned from a list stored the category_names.setter
        '''

    @property
    @abc.abstractmethod
    def num_categories(self):
        '''the number of categories'''


class MultiImageStackClassification(MultiImageStack):
    '''many image stack classifications stored as collections.UserDict'''

    def create_from_multi_image_stack(self, mis: MultiImageStack):
        '''calculate data according to the data shape of MultiImageStack'''


class FormatTrainingData():
    @abc.abstractmethod
    def __init__(self, slice_len, slice_padding):
        '''initialize with slice_len and slice padding arguments'''

    @abc.abstractmethod
    def apply(self, x, y):
        '''slice the numpy arrays in dict x and y into slices and concatenate the slices into a list
        x and y must have identical keys and the first dimension of the numpy arrays must match
        '''


class TrainingDataGenerator(SaveLoad):
    '''generator for training data'''

    @property
    @abc.abstractmethod
    def data_shape(self):
        '''shape of the data that is generated'''

    @property
    @abc.abstractmethod
    def annotation_shape(self):
        '''shape of the annotations that are generated'''

    @abc.abstractmethod
    def steps_per_epoch(self):
        '''number of steps per epoch (total number of datapoints/batch size'''

    @abc.abstractmethod
    def generator(self):
        '''data generator that returns batches of data and annotations'''

    @abc.abstractmethod
    def validation_generator(self):
        '''data used exclusively for validation'''

    @abc.abstractmethod
    def validation_steps(self):
        '''number of steps per epoch (total number of datapoints/batch size'''


class NetworkList(SaveLoad, IsEmpty):
    pass


class NetworkTrainer(SaveLoad):
    def __init__(self, model_factory=None, training_config=None):
        pass

    @abc.abstractmethod
    def compile_model(self):
        '''use model_factory to create and compile a keras model'''

    @abc.abstractmethod
    def draw_model(self, file_name):
        '''draw the model'''

    @abc.abstractmethod
    def train(self):
        '''train the model'''

    @abc.abstractmethod
    def get_fig(self):
        '''get the figure handle for the loss plot'''

    @abc.abstractmethod
    def get_ax(self):
        '''get the axis handle for the loss plot'''

    @abc.abstractmethod
    def set_canvas(self, canvas):
        '''set the canvas for the loss plot'''


class SequenceModel(SaveLoad):

    @abc.abstractmethod
    def configure(self, config: Config):
        '''apply configuration settings'''

    @abc.abstractmethod
    def calc_input_shape(self, data_shape: tuple) -> tuple:
        '''calculate input shape from data shape'''

    @abc.abstractmethod
    def calc_num_categories(self, annotation_shape: tuple) -> int:
        '''calculate num_categories from annotation shape'''

    @abc.abstractmethod
    def compile(self, input_shape, data_shape, output_classes):
        '''assemble and compile the model.

        input shape is either given directly or calculated from data shape
        '''


class PredictionFormatter():
    '''Format data so that it can be passed to SequenceModel.predict and reverse-format predictions'''
    @abc.abstractmethod
    def __init__(self, sequence_len=64, time_overlap=16):
        '''Arguments:
        sequence_len: int, total length of sequences
        time_overlap: int, number of frames that individual sequences should overlap
        '''

    @abc.abstractmethod
    def apply(self, data: MultiImageStack):
        '''Split all stacks in data into substacks of length sequence_len.

        Arguments:
        data: interface.MultiImageStack, data to be split

        Returns:
        numpy.ndarray split stacks
        '''

    @abc.abstractmethod
    def revert(self, data: MultiImageStack, y_pred):
        '''concatenate y_pred into stacks that match the shape of the stacks in data

        Arguments:
        data: MultiImageStack, original data used as template for predicted data
        y_pred: numpy.ndarray, prediction results

        Returns:
        MultiImageStackClassification
        '''


class WorkflowData(metaclass=abc.ABCMeta):
    '''Stores all workflow data as properties.'''

    @property
    @abc.abstractmethod
    def path(self):
        '''path where the data will be stored'''


class VidtrainDataLossException(Exception):
    pass



import collections
from .. import gui
from .. import datahandler
from . import preparation
from . import annotation
from . import training
from . import prediction
from . import interface
from . import evaluation


class JunctionAnalysis():
    def __init__(self, path=None):
        self.root = gui.backend.RootWindow()
        self.root.wm_title('Junction Analysis')
        self.tabs = gui.backend.Tabs(master=self.root)
        self.filedialog = gui.backend.Filedialog()
        self.path = path or self.filedialog.ask_open([])
        self.workflow_data = datahandler.JunctionAnalysisData(path=self.path)
        self.prepare = preparation.JunctionExtraction(self.root, workflow_data=self.workflow_data)
        self.annotate = annotation.JunctionAnnotation(self.root, workflow_data=self.workflow_data)
        self.annotate.autosave = True
        self.train = training.JunctionNetworkTraining(self.root, workflow_data=self.workflow_data)
        self.predict = prediction.JunctionPrediction(self.root, workflow_data=self.workflow_data)
        self.evaluate = evaluation.JunctionEvaluation(self.root, workflow_data=self.workflow_data)
        self.workflow = WorkflowList([self.prepare, self.annotate, self.train, self.predict, self.evaluate])
        self.tabs.extend(self.workflow)
        self.tabs.show(side=gui.interface.TOP, fill=gui.interface.BOTH,
                       expand=gui.interface.ON)
        self.workflow.add_controls(self.root)
        self.workflow.on_step_change = self.tabs.select
        self.tabs.on_tab_select = self.workflow.select_step
        # TODO: (prio 4) check workflowStep.is_done() before activating the next button
        # TODO: (prio 1): add "loading data..." progress bar to the root window while data is being loaded

    def run(self):
        self.root.run()


class WorkflowList(collections.abc.MutableSequence):
    '''A list of workflow steps where the output of one step is passed to the next step as input
    each item is a tab in a ttk.Notebook
    below the tabs is a list of buttons that handles navigation between steps as well as loading input data and saving output data
    '''

    def __init__(self, initlist=None, on_step_change=gui.interface._empty_callback):
        self._cur = 0
        self.on_step_change = on_step_change
        self.data = []
        if initlist is not None:
            self.extend(initlist)

    @property
    def current_step(self):
        return self._cur

    @current_step.setter
    def current_step(self, tab_no):
        self._change_step(tab_no)
        self._cur = tab_no

    def add_controls(self, master):
        self.buttons = gui.backend.Panel(master=master)
        self.button_back = gui.backend.Button(master=self.buttons, text='<< back', on_click=self.back)
        self.button_back.show()
        self.button_load = gui.backend.Button(master=self.buttons, text='load', on_click=self.load)
        self.button_load.show()
        self.button_save = gui.backend.Button(master=self.buttons, text='save', on_click=self.save)
        self.button_save.show()
        self.button_next = gui.backend.Button(master=self.buttons, text='next >>', on_click=self.next)
        self.button_next.show()
        self.buttons.show(side=gui.interface.BOTTOM)

    def __getitem__(self, i): return self.data[i]

    def __delitem__(self, i): del self.data[i]

    def __len__(self): return len(self.data)

    def __setitem__(self, i, v):
        self.check(v)
        self.data[i] = v

    def insert(self, i, v):
        self.check(v)
        self.data.insert(i, v)

    def check(self, v):
        assert isinstance(v, interface.WorkflowStep)
        v.load_input()

    def next(self):
        if self.current_step < len(self) - 1:
            self.current_step += 1

    def back(self):
        if self.current_step > 0:
            self.current_step -= 1

    def load(self):
        self.data[self.current_step].load_input()

    def save(self):
        self.data[self.current_step].save_output()

    def select_step(self, step_no):
        self.current_step = step_no

    def _change_step(self, new_step):
        if new_step != self.current_step:
            self.data[new_step].update()
            self.on_step_change(new_step)



import warnings
import glob
import gzip
import pickle
from time import strftime
from tensorflow.keras.layers import (Input, Dropout, ConvLSTM2D, Conv3D, Concatenate, BatchNormalization,
                                     GlobalMaxPooling3D, GlobalMaxPooling2D, MaxPooling3D, TimeDistributed, Dense)
from tensorflow.keras.models import Model
from tensorflow.keras.callbacks import LearningRateScheduler, Callback, History
from sklearn.model_selection import train_test_split
import numpy as np
import matplotlib.pyplot as plt
from . import loss
from . import interface
from .data import FileVersioning


class NetworkTrainer(interface.NetworkTrainer):

    def __init__(self, model_factory=None, training_data=None, epochs=150):
        self.input_shape = None
        self.num_categories = None
        self.epochs = epochs
        self.training_data = training_data
        self.model_factory = model_factory
        self.model = None
        self.history = None
        self.scheduler = StepDecayScheduler()
        self.loss_plot = PlotLosses()
        self.file_versioning = FileVersioning()
        # ToDo: use TqdmCallback to add a gui progress bar

    def compile_model(self):
        assert self.model_factory is not None, 'Could not compile model. Please set model_factory attribute.'
        if self.input_shape is None or self.num_categories is None:
            assert self.training_data is not None, 'Could not determine input shape. Please set input_shape or training_data attributes.'
            self.input_shape = self.model_factory.calc_input_shape(self.training_data.data_shape)
            self.num_categories = self.model_factory.calc_num_categories(self.training_data.annotation_shape)
        self.model = self.model_factory.compile(input_shape=self.input_shape, num_categories=self.num_categories)

    def draw_model(self, file_name):
        assert self.model is not None, 'Please compile the model with compile_model() first.'
        from tensorflow.keras.utils import model_to_dot
        model_to_dot(self.model, dpi=72).write_svgz(file_name, prog=['dot', '-Nfontname=sans-serif'])

    def train(self):
        assert self.model is not None, 'Please compile the model with compile_model() first.'
        assert self.training_data is not None, 'Could not train without data. Please set training_data property.'
        lrate = LearningRateScheduler(self.scheduler.step_decay)
        callbacks = [lrate]
        if self.loss_plot.canvas is not None:
            callbacks.append(self.loss_plot)
        self.history = self.model.fit(self.training_data.generator(),
                                      steps_per_epoch=self.training_data.steps_per_epoch(),
                                      epochs=self.epochs,
                                      callbacks=callbacks,
                                      validation_data=self.training_data.validation_generator(),
                                      validation_steps=self.training_data.validation_steps())

    def get_fig(self):
        if self.loss_plot.fig is None:
            self.loss_plot.create_fig()
        return self.loss_plot.fig

    def get_ax(self):
        if self.loss_plot.ax is None:
            self.loss_plot.create_fig()
        return self.loss_plot.ax

    def set_canvas(self, canvas):
        self.loss_plot.set_canvas(canvas)

    def copy(self):
        copy = NetworkTrainer(model_factory=self.model_factory, training_data=self.training_data, epochs=self.epochs)
        copy.loss_plot = self.loss_plot
        return copy

    def save(self, path_prefix, unique=True):
        path_prefix = self.file_versioning.get_name(base_path=path_prefix, unique=unique)
        params = {'input_shape': self.input_shape,
                  'num_categories': self.num_categories,
                  'factory_type': type(self.model_factory)}
        with gzip.open(path_prefix + '_params.pkl.gz', 'w') as f:
            pickle.dump(params, f)
        try:
            self.model_factory.save(path_prefix + '_model.pkl.gz')
            params['history'] = self.history.history
            with gzip.open(path_prefix + '_params.pkl.gz', 'w') as f:
                pickle.dump(params, f)
            self.model.save_weights(path_prefix + '_weights.h5')
        except AttributeError:
            pass
        return path_prefix

    def load(self, path_prefix):
        with gzip.open(path_prefix + '_params.pkl.gz', 'r') as f:
            params = pickle.load(f)
        self.input_shape = params['input_shape']
        self.num_categories = params['num_categories']
        self.model_factory = params['factory_type']()
        if 'history' in params.keys():
            self.history = History()
            self.history.history = params['history']
        try:
            self.model_factory.load(path_prefix + '_model.pkl.gz')
            self.compile_model()
            self.model.load_weights(path_prefix + '_weights.h5')
        except (AttributeError, FileNotFoundError):
            pass


class SequenceModelMixin(interface.SequenceModel):
    def __init__(self, config=None):
        self.conv3d_layers = 1
        self.convlstm2d_layers = 4
        self.kernels = 32
        self.optimizer = 'adam'
        self.dropout = 0
        self.relu_initializer = 'he_uniform'
        self.sigmoidal_initializer = 'glorot_uniform'
        self.recurrent_initializer = 'orthogonal'
        if config is not None:
            self.configure(config)

    def configure(self, config: interface.Config):
        for config in config.values():
            try:
                value = config.current_option if hasattr(config, 'current_option') else config.value
                setattr(self, config.name2var(), value)
            except AttributeError:
                warnings.warn('did not recognize config option: {}'.format(config.name))

    def _dense_conv_lstm(self, layers, connected_layers=(), conv3d=None, lstm=None) -> tuple:
        conv3d = conv3d or self.conv3d_layers
        lstm = lstm or self.convlstm2d_layers
        for _ in range(conv3d):
            (layers, connected_layers) = self._add_connected_layer(
                layers, connected_layers,
                layer_type='conv3d')
        for _ in range(lstm):
            (layers, connected_layers) = self._add_connected_layer(
                layers, connected_layers,
                layer_type='lstm')
        return (layers, connected_layers)

    def _add_connected_layer(self, layers, connected_layers, layer_type='conv3d') -> tuple:
        if layer_type == 'conv3d':
            temp_l = Conv3D(self.kernels,
                            kernel_size=(3, 3, 3),
                            activation='relu',
                            padding='same',
                            kernel_initializer=self.relu_initializer)(layers)
        else:
            temp_l = ConvLSTM2D(self.kernels,
                                kernel_size=(3, 3),
                                activation='tanh',
                                recurrent_activation='hard_sigmoid',
                                padding='same',
                                return_sequences=True,
                                kernel_initializer=self.sigmoidal_initializer,
                                recurrent_initializer=self.recurrent_initializer)(layers)
        if self.dropout > 0:
            temp_l = Dropout(self.dropout)(temp_l)
        temp_l = BatchNormalization()(temp_l)
        connected_layers += (temp_l,)
        if len(connected_layers) > 1:
            layers = Concatenate()(list(connected_layers))
        else:
            layers = connected_layers[0]
        return (layers, connected_layers)

    def save(self, path):
        with gzip.open(path, 'w') as f:
            pickle.dump(self.__dict__, f)

    def load(self, path):
        with gzip.open(path, 'r') as f:
            loaded = pickle.load(f)
        assert isinstance(loaded, dict), 'file must contain a dict. Got" {}'.format(type(loaded))
        self.__dict__ = loaded


class SequenceSegmentationModel(SequenceModelMixin):

    def calc_input_shape(self, data_shape: tuple) -> tuple:
        '''calculate input shape from data shape'''
        return (None, None, None) + data_shape[3:]

    def calc_num_categories(self, annotation_shape: tuple) -> int:
        '''calculate num_categories from annotation shape'''
        return annotation_shape[3]

    def compile(self, input_shape=(None, None, None, 1), data_shape=None, num_categories=1) -> Model:
        '''assemble and compile the model.

        input shape is either given directly or calculated from data shape
        '''
        if data_shape is not None:
            input_shape = self.calc_input_shape(data_shape)
        i = Input(input_shape)
        layers = i
        layers, connected_layers = self._dense_conv_lstm(
            layers, conv3d=self.conv3d_layers - 1, lstm=self.convlstm2d_layers - 1)
        layers, _ = self._dense_conv_lstm(layers, connected_layers, conv3d=1, lstm=1)
        o = Conv3D(num_categories, kernel_size=(3, 3, 3), activation='sigmoid', padding='same')(layers)
        model = Model(inputs=i, outputs=o)
        model.compile(loss=loss.dice_loss, optimizer=self.optimizer)
        return model


class SequenceClassificationModel(SequenceModelMixin):
    def __init__(self, config=None):
        self.reduction_layers = 3
        self.reduction_layers_step = 1
        self.grow_kernels = 1.0
        self.fully_connected = False
        super().__init__(config=config)

    def calc_input_shape(self, data_shape: tuple) -> tuple:
        '''calculate input shape from data shape'''
        return (data_shape[0], None, None) + data_shape[3:]

    def calc_num_categories(self, annotation_shape: tuple) -> int:
        '''calculate num_categories from annotation shape'''
        return annotation_shape[1]

    def compile(self, input_shape=(64, None, None, 1), data_shape=None, num_categories=6) -> Model:
        '''assemble and compile the model.

        input shape is either given directly or calculated from data shape
        '''
        if data_shape is not None:
            input_shape = self.calc_input_shape(data_shape)
        i = Input(input_shape)
        layers = i
        layers, connected_layers = self._dense_conv_lstm(layers)
        if not self.fully_connected:
            num_kernels = self.kernels * (self.conv3d_layers + self.convlstm2d_layers)
        else:
            num_kernels = self.kernels
        for _ in range(self.reduction_layers):
            (layers, connected_layers, num_kernels) = self._add_reduction_layer(
                layers,
                connected_layers,
                num_kernels)
        if self.dropout > 0:
            layers = Dropout(self.dropout)(layers)
        layers = BatchNormalization()(layers)
        if self.fully_connected:
            connected_layers += (layers,)
            c_l = ()
            for l in connected_layers:
                c_l += (TimeDistributed(GlobalMaxPooling2D())(l),)
            layers = Concatenate()(list(c_l))
        else:
            layers = TimeDistributed(GlobalMaxPooling2D())(layers)
        o = TimeDistributed(Dense(num_categories,
                                  activation='sigmoid',
                                  kernel_initializer=self.sigmoidal_initializer))(layers)
        model = Model(inputs=i, outputs=o)
        model.compile(loss=loss.dice_loss, optimizer=self.optimizer)
        return model

    def _add_reduction_layer(self, layers, connected_layers, num_kernels) -> tuple:
        layers = MaxPooling3D(pool_size=(1, 2, 2))(layers)
        num_kernels = int(num_kernels * self.grow_kernels)
        for _ in range(self.reduction_layers_step):
            if self.dropout > 0:
                layers = Dropout(self.dropout)(layers)
            layers = BatchNormalization()(layers)
            if self.fully_connected:
                connected_layers += (layers,)
            layers = Conv3D(num_kernels,
                            (3, 3, 3),
                            activation='relu',
                            padding='same',
                            kernel_initializer=self.relu_initializer)(layers)
        return (layers, connected_layers, num_kernels)


class StepDecayScheduler:
    def __init__(self, initial_lrate=0.002, drop=0.5, epochs_drop=8.0):
        self.initial_lrate = initial_lrate
        self.drop = drop
        self.epochs_drop = epochs_drop

    def step_decay(self, epoch) -> float:
        lrate = self.initial_lrate * np.power(self.drop,
                                              np.floor((1 + epoch) / self.epochs_drop))
        return lrate


class PlotLosses(Callback):
    def __init__(self):
        Callback.__init__(self)
        self.fig = None
        self.ax = None
        self.canvas = None

    def create_fig(self):
        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(111)

    def set_canvas(self, canvas):
        self.canvas = canvas
        self.canvas.update()
        self.ax.clear()
        self.ax.text(0.25, 0.5, 'starting training, please wait...')
        self.canvas.update()

    def on_train_begin(self, logs={}):
        self.i = 1
        self.x = []
        self.losses = []
        self.val_losses = []
        self.logs = []

    def on_epoch_end(self, epoch, logs={}):
        self.logs.append(logs)
        self.x.append(self.i)
        self.losses.append(logs.get('loss'))
        self.val_losses.append(logs.get('val_loss'))
        self.i += 1
        self.ax.clear()
        self.ax.plot(self.x, self.losses, label="loss", figure=self.fig)
        self.ax.plot(self.x, self.val_losses, label="val_loss", figure=self.fig)
        self.ax.legend()
        plt.xlabel('epochs')
        plt.ylabel('loss')
        self.canvas.update()


class TrainingDataGenerator(interface.TrainingDataGenerator):
    '''generator for training data'''

    def __init__(self, x=None, y=None, batch_size=8, len_sequences=64, test_size=0.2):
        self.batch_size = batch_size
        self.len_sequences = len_sequences
        self.test_size = test_size
        self.random_state = 42
        self.x_train = self.x_test = self.y_train = self.y_test = None
        if x is not None:
            self.set_data(x, y)

    def set_data(self, x: list, y: list):
        self.x_train, self.x_test, self.y_train, self.y_test = train_test_split(
            x, y, test_size=self.test_size, random_state=self.random_state)

    @property
    def data_shape(self):
        '''shape of the data that is generated'''
        return (self.len_sequences,) + self.x_train[0].shape[1:]

    @property
    def annotation_shape(self):
        '''shape of the annotations that are generated'''
        return (self.len_sequences,) + self.y_train[0].shape[1:]

    def steps_per_epoch(self):
        '''number of steps per epoch (total number of datapoints/batch size'''
        return int(sum([x.shape[0] for x in self.x_train]) / self.len_sequences / self.batch_size)

    def generator(self, train=True, aug_shift=0):
        '''data generator that returns batches of data and annotations'''
        if train:
            x = self.x_train
            y = self.y_train
        else:
            x = self.x_test
            y = self.y_test
        indices = list(range(len(x)))
        np.random.shuffle(indices)
        sample_no = 0
        while True:
            stacks = []
            labels = []
            while len(stacks) < self.batch_size:
                stack = np.zeros((0,) + x[0].shape[1:])
                label = np.zeros((0,) + y[0].shape[1:])
                while stack.shape[0] < self.len_sequences:
                    if sample_no > len(indices) - 1:
                        np.random.shuffle(indices)
                        sample_no = 0
                    sample = x[indices[sample_no]]
                    missing_frames = self.len_sequences - stack.shape[0]
                    if sample.shape[0] <= missing_frames:
                        stack = np.vstack((stack, sample))
                        label = np.vstack((label, y[indices[sample_no]]))
                    else:
                        stack = np.vstack((stack, sample[:missing_frames, ]))
                        label = np.vstack((label, y[indices[sample_no]][:missing_frames, ]))
                    sample_no += 1
                if aug_shift > 0:
                    from scipy.ndimage import shift
                    dx = np.random.randint(-aug_shift, aug_shift)
                    dy = np.random.randint(-aug_shift, aug_shift)
                    stack = shift(stack, (0, dx, dy, 0), order=0, mode='nearest')
                # TODO: (prio 3) add augmentation for junction image size and framerate
                stacks.append(stack)
                labels.append(label)
            yield (np.array(stacks), np.array(labels))

    def validation_generator(self):
        '''data used exclusively for validation'''
        return self.generator(train=False)

    def validation_steps(self):
        '''number of steps per epoch (total number of datapoints/batch size'''
        return int(sum([x.shape[0] for x in self.x_test]) / self.len_sequences / self.batch_size)

    def save(self, path_prefix):
        np.savez_compressed(path_prefix + '_data.npz',
                            x_train=np.array(self.x_train), y_train=np.array(self.y_train),
                            x_test=np.array(self.x_test), y_test=np.array(self.y_test))
        params = {'batch_size': self.batch_size, 'len_sequences': self.len_sequences}
        with gzip.open(path_prefix + '_params.pkl.gz', 'w') as f:
            pickle.dump(params, f)

    def load(self, path_prefix):
        with np.load(path_prefix + '_data.npz') as data:
            self.x_train = list(data['x_train'])
            self.y_train = list(data['y_train'])
            self.x_test = list(data['x_test'])
            self.y_test = list(data['y_test'])
        with gzip.open(path_prefix + '_params.pkl.gz', 'r') as f:
            params = pickle.load(f)
        self.batch_size = params['batch_size']
        self.len_sequences = params['len_sequences']



import os
import numpy as np
from matplotlib import pyplot as plt
from .. import gui
from .. import datahandler
from . import interface


class JunctionEvaluation(gui.backend.Panel, interface.WorkflowStep):
    NAME = 'Evaluate'

    def __init__(self, master, workflow_data: datahandler.interface.WorkflowData):
        gui.backend.Panel.__init__(self, master=master)
        self.workflow_data = workflow_data
        self.file_versioning = datahandler.FileVersioning()
        self.current_stack = 0
        self.config = datahandler.Config(panel_name='')
        self.config.new_item(name='Result type', value=0, options=self._get_options(), on_change=self.update_everything)
        self.config_panel = gui.backend.ConfigPanel(master=self, config=self.config)
        self.config_panel.show(side=gui.interface.TOP, pady=5)
        self.single_stack_container = gui.backend.Panel(master=self)
        self.single_stack_container.add_label('Individual Stack Results')
        self.single_plots = gui.backend.ClassificationEvaluationPlots(master=self.single_stack_container)
        self.controls = gui.backend.StackControls(master=self.single_stack_container, callback=self.update)
        self.single_plots.show(side=gui.interface.TOP)
        self.controls.show(side=gui.interface.TOP, fill=gui.interface.X, expand=gui.interface.ON)
        self.single_stack_container.show(side=gui.interface.TOP, pady=5)
        self.combined_container = gui.backend.Panel(master=self)
        self.combined_container.add_label('Combined Results')
        self.combined_classes = gui.backend.ClassificationEvalPlot(
            master=self.combined_container, fig=plt.figure(), height=384, aspect=0.5)
        self.combined_classes.add_label('Combined Classes')
        self.combined_junctions = gui.backend.ClassificationEvalPlot(
            master=self.combined_container, fig=plt.figure(), height=384, plot='combined', aspect=2)
        self.combined_junctions.add_label('Total events per junction')
        self.combined_classes.show(side=gui.interface.LEFT)
        self.combined_junctions.show(side=gui.interface.LEFT)
        self.combined_container.show(side=gui.interface.TOP, pady=5)
        self.force_update = False
        self.update()
        # TODO: (prio 2) configure junction types

    @property
    def input(self):
        '''Input data to be processed'''
        return (self.workflow_data.annotation, self.workflow_data.pred_med, self.workflow_data.pred_std, self.workflow_data.pred_corr)

    @property
    def output(self):
        '''Output data after processing'''
        return None

    def update(self, _=None):
        if self.is_ready():
            self.config['Result type'].options = self._get_options()
            if self.config['Result type'].current_option == 'annotations':
                data = self.workflow_data.annotation
            elif self.config['Result type'].current_option == 'predictions':
                data = self.workflow_data.pred_med
            elif self.config['Result type'].current_option == 'corrected predictions':
                data = self.workflow_data.pred_corr
            self.controls.max_frame = len(data) - 1
            self.single_plots.update(data.image_stacks(self.controls.frame))
            if self.combined_classes.classification is None or self.force_update:
                self.update_combined_plots(data)

    def update_combined_plots(self, data):
        classes = datahandler.ImageStackClassification(
            data=np.concatenate(data.np_array()),
            category_names=list(data.category_names))
        self.combined_classes.update(classes)
        junctions = data.np_array()
        if len(junctions.shape) > 1:
            junctions = np.sum(junctions, axis=tuple(range(1, junctions.ndim)))[..., np.newaxis]
        else:
            js = np.zeros(junctions.shape)
            for i, j in enumerate(junctions):
                js[i] = np.sum(j)
            junctions = js[..., np.newaxis]
        self.combined_junctions.update(datahandler.ImageStackClassification(
            data=junctions, category_names=data.names))

    def update_everything(self, _=None):
        self.force_update = True
        self.update()

    def is_ready(self):
        '''Check if this step is ready to be run.'''
        return not self.workflow_data.annotation.empty()

    def is_done(self):
        '''Check if this step is done and ready for the next step.'''
        return not self.workflow_data.annotation.empty()

    def load_input(self):
        '''Ask user for data path and load input.'''
        if self.workflow_data.path is None:
            fd = gui.backend.Filedialog()
            self.workflow_data.path = fd.ask_open(filetypes=([]))
        path = self.default_in_file()
        try:
            self.workflow_data.annotation.load(self.file_versioning.get_newest_file(path[0]))
            self.workflow_data.pred_med.load(self.file_versioning.get_newest_file(path[1]))
            self.workflow_data.pred_std.load(self.file_versioning.get_newest_file(path[2]))
            self.workflow_data.pred_corr.load(self.file_versioning.get_newest_file(path[3]))
        except TypeError:
            pass
        self.update()

    def save_output(self):
        '''Ask user for data path and save output.'''
        pass

    def default_in_file(self):
        '''Calculate default input file path relative to path.
        Arguments:
        path: string path of original file
        Returns:
        string: default output file path
        '''
        paths = (os.path.join(datahandler.get_eval_dir(self.workflow_data.path), 'junction_annotations.npz'),
                 os.path.join(datahandler.get_eval_dir(self.workflow_data.path), 'junction_predicted_mean.npz'),
                 os.path.join(datahandler.get_eval_dir(self.workflow_data.path), 'junction_predicted_std.npz'),
                 os.path.join(datahandler.get_eval_dir(self.workflow_data.path), 'junction_predicted_manual_corrected.npz'))
        return paths

    def default_out_file(self):
        '''Calculate default output file path relative to path.
        Arguments:
        path: string path of original file
        Returns:
        string: default output file path
        '''
        pass

    def _get_options(self):
        options = ['annotations']
        if not self.workflow_data.pred_med.empty():
            options.append('predictions')
        if not self.workflow_data.pred_corr.empty():
            options.append('corrected predictions')
        return options



import os
import numpy as np
import micdata
from .. import gui
from .. import datahandler
from . import interface


class JunctionExtraction(gui.backend.Panel, interface.WorkflowStep):
    NAME = 'Extract Junctions'

    def __init__(self, master, workflow_data: datahandler.interface.WorkflowData):
        gui.backend.Panel.__init__(self, master=master)
        self.workflow_data = workflow_data
        self.projection_config = datahandler.Config()
        self.projection_config.new_item(name='Projection method', value=0, options=['median', 'std', 'max', 'mean'])
        self.projection_chooser = gui.backend.Panel(master=self)
        self.proj_config_panel = gui.backend.ConfigPanel(master=self.projection_chooser, config=self.projection_config)
        self.proj_config_panel.show(side=gui.interface.LEFT)
        self.create_image_btn = gui.backend.Button(
            self.projection_chooser, 'apply', on_click=self.re_calculate_image)
        self.create_image_btn.show(side=gui.interface.LEFT, padx=5, pady=2)
        self.projection_chooser.show(side=gui.interface.TOP, fill=gui.interface.X, expand=gui.interface.ON)
        self.input_cache = None
        self.image = None
        self.rectangles = None
        self.main_figure = None
        self.main_figure_panel = None
        self.config_panel = None
        self.config = datahandler.Config(panel_name='Rectangle config:')
        self.config.new_item(name='Rows', value=1,
                             from_=1, to=100, on_change=self._on_row_update)
        self.config.new_item(name='Columns', value=16,
                             from_=2, to=100, on_change=self._on_col_update)
        self.config.new_item(name='Rectangle size (px)', value=16,
                             from_=8, to=256, increment=8, on_change=self._on_dim_update)
        self.config.new_item(name='Row height (px)', value=16.0,
                             from_=4.0, to=64.0, increment=0.01, on_change=self._on_height_update)
        self.config.new_item(name='Move', value=0, options=[
                             'red rectangle', 'blue rectangle'], on_change=self._select_rectangle)
        self.buttons = gui.backend.Panel(master=self)
        self.done_button = gui.backend.Button(master=self.buttons, text='extract junctions')
        self.done_button.show(side=gui.interface.LEFT)
        self.buttons.show(side=gui.interface.BOTTOM)
        self.dialogs = gui.backend.Dialogs(title='Junction Extraction')
        # TODO: (prio 4) save/load rectangles and median image
        # TODO: (prio 5) add optional background subtraction

    @property
    def input(self):
        '''Input data to be processed'''
        return self.workflow_data.raw

    @input.setter
    def input(self, input_data: np.ndarray):
        self.workflow_data.raw.data = input_data
        self.update()

    @property
    def config(self):
        '''Configuration data'''
        return self._config

    @config.setter
    def config(self, conf: datahandler.interface.Config):
        self._config = conf

    @property
    def output(self):
        '''Output data after processing'''
        return self.workflow_data.processed

    @output.setter
    def output(self, output_data: datahandler.interface.MultiImageStack):
        self.workflow_data.processed.update(output_data)

    def figure_panel_label(self):
        return self.projection_config['Projection method'].current_option + ' projection:'

    def cache_data(self):
        self.input_cache = np.copy(self.input.data)

    def data_changed(self):
        return not np.array_equal(self.input_cache, self.input.data)

    def create_image(self):
        if self.data_changed():
            self.image = getattr(self.input, self.projection_config['Projection method'].current_option)()
            self.cache_data()

    def re_calculate_image(self):
        self.input_cache = None
        self.main_figure_panel.edit_label(self.figure_panel_label())
        self.update()

    def update(self):
        if self.is_ready():
            self.create_image()
            try:
                self.main_figure.update(self.image)
            except AttributeError:
                self.create_gui()

    def create_gui(self):
        self.rectangles = datahandler.rotatable.RectangleMatrix(
            image_shape=(self.input.image_shape[:2]),
            num_rows=self.config['Rows'].value,
            num_cols=self.config['Columns'].value,
            size=self.config['Rectangle size (px)'].value,
            row_height=self.config['Row height (px)'].value)
        self.fig_zoom_canvas = gui.backend.Panel(master=self)
        self.main_figure_panel = gui.backend.Panel(master=self.fig_zoom_canvas)
        self.main_figure_panel.add_label(self.figure_panel_label())
        self.main_figure = gui.backend.RotatableImagePanel(master=self.main_figure_panel, image=self.image)
        self.main_figure.add_controls()
        self.main_figure.show()
        self.main_figure_panel.show(padx=2)
        self.rectangles.add_all(self.main_figure.get_ax())
        self.main_figure_mouse = gui.backend.FigurePanelMouse(
            figure_canvas=self.main_figure,
            press_callback=self._on_mouse_button_press,
            drag_callback=self._on_mouse_drag)
        self.main_figure_key = gui.backend.FigurePanelKey(
            self.main_figure,
            callback=self._on_key_press)
        self.main_figure.rotation_callback = self._inc_rotation
        self.main_figure.flip_callback = self._toggle_flip
        self.zoom = gui.backend.Panel(self.fig_zoom_canvas)
        self.zoom.add_label(text='left')
        self.left_zoom_figure = gui.backend.RotatableImagePanel(
            master=self.zoom, image=self._extract_rect_image(self.rectangles.left_rect), toolbar=False, zoom=self.image.shape[1] / 2 / self.config['Rectangle size (px)'].value)
        self.left_zoom_figure.show(side=gui.interface.TOP)
        self.zoom.add_label(text='right')
        self.right_zoom_figure = gui.backend.RotatableImagePanel(
            master=self.zoom, image=self._extract_rect_image(self.rectangles.right_rect), toolbar=False, zoom=self.image.shape[1] / 2 / self.config['Rectangle size (px)'].value)
        self.right_zoom_figure.show(side=gui.interface.TOP)
        self.zoom.show(side=gui.interface.TOP)
        self.fig_zoom_canvas.show(side=gui.interface.TOP)
        self.config_panel = gui.backend.ConfigPanel(master=self, config=self.config)
        self.config_panel.show(side=gui.interface.BOTTOM, fill=gui.interface.X, expand=gui.interface.ON)
        self.done_button.on_click = self._save_rect_stacks

    def is_ready(self):
        '''Check if this step is ready to be run.'''
        return not self.input.empty()

    def is_done(self):
        '''Check if this step is done and the next step can be run.'''
        return isinstance(self.output, datahandler.interface.MultiImageStack)

    def load_input(self):
        self.input.load(self.default_in_file())
        self.update()
        if os.path.exists(self.default_out_file()):
            self.output.load(self.default_out_file())

    def save_output(self):
        if self.workflow_data.path is None:
            fd = gui.backend.Filedialog()
            self.workflow_data.path = fd.ask_save(filetypes=([('numpy compressed', 'npz')]))
        self.output.save(self.default_out_file())

    def default_in_file(self):
        return self.workflow_data.path

    def default_out_file(self):
        return os.path.join(datahandler.get_eval_dir(self.workflow_data.path), 'junction_stacks.npz')

    def _inc_rotation(self):
        self.rectangles.rotate()
        self._update_fig()

    def _toggle_flip(self):
        self.rectangles.flip()
        self._update_fig()

    def _on_row_update(self, rows):
        self.rectangles.num_rows = int(rows)
        self.rectangles.update_rect_number()
        self.rectangles.refresh_ax(self.main_figure.im.ax)
        self.main_figure.figure.draw()

    def _on_col_update(self, cols):
        self.rectangles.num_cols = int(cols)
        self.rectangles.update_rect_number()
        self.rectangles.refresh_ax(self.main_figure.im.ax)
        self.main_figure.figure.draw()

    def _on_dim_update(self, size):
        self.rectangles.dim = int(size)
        self._update_fig()

    def _on_height_update(self, height):
        self.rectangles.row_height = height
        self._update_fig()

    def _update_fig(self):
        self.rectangles.update()
        self.main_figure.update()
        self._update_zoom()

    def _update_zoom(self):
        self.left_zoom_figure.update(image=self._extract_rect_image(self.rectangles.left_rect))
        self.right_zoom_figure.update(image=self._extract_rect_image(self.rectangles.right_rect))

    def _on_mouse_button_press(self, event):
        if event.button == 1:
            self.rectangles.store_current_pos((event.xdata, event.ydata))
            self.main_figure.update()
            self._update_zoom()
        else:
            self.config['Move'].toggle()

    def _on_mouse_drag(self, event):
        self.rectangles.store_current_pos((event.xdata, event.ydata))
        self.main_figure.update()
        self._update_zoom()

    def _on_key_press(self, key):
        self.rectangles.move_current_rect(key)
        self.main_figure.update()
        self._update_zoom()

    def _select_rectangle(self, rect):
        if rect == 0:
            self.rectangles.select_left()
        else:
            self.rectangles.select_right()

    def _save_rect_stacks(self):
        if len(self.output) > 0:
            dialog = gui.backend.Dialogs(title='append data?')
            if not dialog.yes_no('there is already extracted junction data. Should the data be appended? (existing data will be lost if you answer "no"!)'):
                self.output.clear()
        for rect in self.rectangles.generator():
            pos = list(map(int, rect.raw_pos))
            if any(p < 0 for p in pos):
                continue
            dim = rect.dim
            try:
                substack = self.input.copy_tile(pos, dim)
                substack = self.main_figure.im.rotate_stack(substack)
            except IndexError:
                continue
            self.output.append(substack, pos)
        self.dialogs.info(message='Extracted {} junction stacks in total.'.format(len(self.output)))

    def _extract_rect_image(self, rect):
        return rect.extract_image(self.main_figure.im)



import tensorflow as tf


def dice_loss(y_true, y_pred):
    return 1-dice_coe(y_true, y_pred, smooth=1)


def dice_coe(output, target, loss_type='sorensen', axis=(1, 2), smooth=1e-5):
    """Soft dice (Sørensen or Jaccard) coefficient for comparing the similarity
    of two batches of data, usually used for binary image segmentation
    i.e. labels are binary. The coefficient ranges between 0 and 1; 1 means complete match.
    from: https://tensorlayer.readthedocs.io/en/latest/_modules/tensorlayer/cost.html
    Apache 2.0 license

    Parameters
    -----------
    output : Tensor
        A distribution with shape: [batch_size, ....], (any dimensions).
    target : Tensor
        The target distribution, format the same with `output`.
    loss_type : str
        ``jaccard`` or ``sorensen``, default is ``jaccard``.
    axis : tuple of int
        All dimensions are reduced, default ``[1,2,3]``.
    smooth : float
        This small value will be added to the numerator and denominator.
            - If both output and target are empty, it makes sure dice is 1.
            - If either output or target are empty (all pixels are background), dice = ```smooth/(small_value + smooth)``, then if smooth is very small, dice close to 0 (even the image values lower than the threshold), so in this case, higher smooth can have a higher dice.

    References
    -----------
    - `Wiki-Dice <https://en.wikipedia.org/wiki/Sørensen–Dice_coefficient>`__

    """
    inse = tf.reduce_sum(output * target, axis=axis)
    if loss_type == 'jaccard':
        l = tf.reduce_sum(output * output, axis=axis)
        r = tf.reduce_sum(target * target, axis=axis)
    elif loss_type == 'sorensen':
        l = tf.reduce_sum(output, axis=axis)
        r = tf.reduce_sum(target, axis=axis)
    else:
        raise Exception("Unknow loss_type")
    dice = (2. * inse + smooth) / (l + r + smooth)
    dice = tf.reduce_mean(dice, name='dice_coe')
    return dice


def dice_hard_coe(output, target, threshold=0.5, axis=(1, 2, 3), smooth=1e-5):
    """Non-differentiable Sørensen–Dice coefficient for comparing the similarity
    of two batches of data. Applies a threshold to the output data before comparing 
    it to target. Usually used for binary image segmentation i.e. labels are binary.
    The coefficient ranges between 0 and 1; 1 means complete match.
    from: https://tensorlayer.readthedocs.io/en/latest/_modules/tensorlayer/cost.html
    Apache 2.0 license

    Parameters
    -----------
    output : tensor
        A distribution with shape: [batch_size, ....], (any dimensions).
    target : tensor
        The target distribution, format the same with `output`.
    threshold : float
        The threshold value to be true.
    axis : tuple of integer
        All dimensions are reduced, default ``(1,2,3)``.
    smooth : float
        This small value will be added to the numerator and denominator, see ``dice_coe``.

    References
    -----------
    - `Wiki-Dice <https://en.wikipedia.org/wiki/Sørensen–Dice_coefficient>`__

    """
    output = tf.cast(output > threshold, dtype=tf.float32)
    target = tf.cast(target > threshold, dtype=tf.float32)
    inse = tf.reduce_sum(tf.multiply(output, target), axis=axis)
    l = tf.reduce_sum(output, axis=axis)
    r = tf.reduce_sum(target, axis=axis)
    hard_dice = (2. * inse + smooth) / (l + r + smooth)
    hard_dice = tf.reduce_mean(hard_dice, name='hard_dice')
    return hard_dice



import os
from .. import gui
from .. import datahandler
from . import interface


class JunctionPrediction(gui.backend.Panel, interface.WorkflowStep):
    NAME = 'Test/Predict'

    def __init__(self, master, workflow_data: datahandler.interface.WorkflowData):
        gui.backend.Panel.__init__(self, master=master)
        self.workflow_data = workflow_data
        self.network_list_panel = None
        self.display_height = 384
        self.input_panel = gui.backend.Panel(master=self)
        self.buttons = gui.backend.Panel(master=self)
        self.load_external_button = gui.backend.Button(
            master=self.buttons, text='load networks', on_click=self.load_external_networks)
        self.load_external_button.show()
        self.prediction_button = gui.backend.Button(master=self.buttons, text='start prediction')
        self.prediction_button.show(side=gui.interface.LEFT)
        self.buttons.show(side=gui.interface.BOTTOM)
        self.sequence_len = 64
        self.time_overlap = 16  # TODO: (prio 2) get this from the network panel
        self.file_versioning = datahandler.FileVersioning()
        if self.is_ready():
            self.update()

    @property
    def input(self):
        '''Input data to be processed'''
        return (self.workflow_data.networks, self.workflow_data.processed, self.workflow_data.annotation)

    @input.setter
    def input(self, inp: tuple):
        assert isinstance(inp[0], datahandler.interface.NetworkList)
        self.workflow_data.networks.update(inp[0])
        assert isinstance(inp[1], datahandler.interface.MultiImageStack)
        self.workflow_data.processed.update(inp[1])
        if len(inp) > 2:
            assert isinstance(inp[2], datahandler.interface.MultiImageStackClassification)
            self.workflow_data.annotation.update(inp[2])
        self.update()

    @property
    def config(self):
        '''Configuration data'''
        return self._config

    @config.setter
    def config(self, conf: datahandler.interface.Config):
        self._config = conf

    @property
    def output(self):
        '''Output data after processing'''
        return (self.workflow_data.pred_med, self.workflow_data.pred_std, self.workflow_data.pred_corr)

    @output.setter
    def output(self, output_data:  tuple):
        for item in output_data:
            assert isinstance(item, datahandler.interface.MultiImageStackClassification)
        self.workflow_data.pred_med.update_from(output_data[0])
        if len(output_data) > 1:
            self.workflow_data.pred_std.update_from(output_data[1])
        if len(output_data) > 2:
            self.workflow_data.pred_corr.update_from(output_data[2])

    def update(self):
        if self.is_ready():
            try:
                self.data_panel.update(self.workflow_data.processed)
                self.network_list_panel.destroy()
            except AttributeError:
                self.create_gui()
            self._add_networklist_panel()
            self.sequence_len = self.workflow_data.networks.get_network(0).input_shape[0]
            self.data_panel.annotation.force = True
            self.data_panel.num_categories = self.workflow_data.networks.get_network(0).num_categories

    def create_gui(self):
        self.output = (datahandler.MultiImageStackClassification(multi_image_stack=self.workflow_data.processed),)
        self.network_list_placeholder = gui.backend.Panel(master=self.input_panel)
        self.network_list_placeholder.show(side=gui.interface.LEFT)
        self.data_panel = gui.backend.MultiStacksAnnotations(
            master=self.input_panel, multi_stack=self.workflow_data.processed)
        # TODO: show current value under mouse
        self.data_panel.show(side=gui.interface.LEFT)
        self.input_panel.show(side=gui.interface.TOP)
        self.prediction_button.on_click = self.start_prediction

    def _add_networklist_panel(self):
        self.network_list_panel = gui.backend.NetworkListPanel(
            master=self.network_list_placeholder, network_list=self.workflow_data.networks)
        self.network_list_panel.show(side=gui.interface.LEFT)

    def is_ready(self):
        '''Check if this step is ready to be run.'''
        return not self.workflow_data.processed.empty() and not self.workflow_data.networks.empty()

    def is_done(self):
        '''Check if this step is done and ready for the next step.'''
        return not self.workflow_data.pred_med.empty()

    def load_input(self):
        if self.workflow_data.path is None:
            fd = gui.backend.Filedialog()
            self.workflow_data.path = fd.ask_open(filetypes=([]))
        path = self.default_in_file()
        self.workflow_data.networks.load(path[0])
        self.update()

    def save_output(self):
        paths = self.default_out_file()
        self.workflow_data.pred_med.save(paths[0])
        self.workflow_data.pred_std.save(paths[1])
        self.data_panel.annotation.save(paths[2])

    def default_in_file(self):
        return (os.path.join(datahandler.get_eval_dir(self.workflow_data.path), 'network'), os.path.join(datahandler.get_eval_dir(self.workflow_data.path), 'junction_stacks.npz'), os.path.join(datahandler.get_eval_dir(self.workflow_data.path), 'junction_annotations.npz'))

    def default_out_file(self):
        paths = (os.path.join(datahandler.get_eval_dir(self.workflow_data.path), 'junction_predicted_mean.npz'),
                 os.path.join(datahandler.get_eval_dir(self.workflow_data.path), 'junction_predicted_std.npz'),
                 os.path.join(datahandler.get_eval_dir(self.workflow_data.path), 'junction_predicted_manual_corrected.npz'))
        return paths

    def start_prediction(self, formatter_class=datahandler.SequencePredictionFormatter):
        formatter = formatter_class(sequence_len=self.sequence_len, time_overlap=self.time_overlap)
        self.workflow_data.pred_med, self.workflow_data.pred_std = self.workflow_data.networks.predict(
            self.workflow_data.processed, formatter)
        self.workflow_data.pred_corr = self.workflow_data.pred_med.copy()
        self.data_panel.annotation = self.workflow_data.pred_corr
        self.data_panel.numpaths_update(self.workflow_data.pred_med.num_categories)

    def load_external_networks(self):
        fd = gui.backend.Filedialog()
        path = fd.ask_open(filetypes=([('taining weights', '.h5')]))
        try:
            path = self.file_versioning.get_part_before_time(path)
            if len(self.file_versioning.get_file_list(path)) > 0:
                self.workflow_data.networks.clear()
                self.workflow_data.networks.load(path)
                self.update()
        except TypeError:
            pass



import os
from .. import gui
from .. import datahandler
from . import interface


class JunctionNetworkTraining(gui.backend.Panel, interface.WorkflowStep):
    NAME = 'Train Network'

    def __init__(self, master, workflow_data: datahandler.interface.WorkflowData):
        gui.backend.Panel.__init__(self, master=master)
        self.workflow_data = workflow_data
        self.fig_panel = None
        self.file_versioning = datahandler.FileVersioning()
        self.input_config = datahandler.Config(panel_name='Input configuration:')
        self.input_config.new_item(name='Slice length', value=64, from_=16, to=128, increment=16)
        self.input_config.new_item(name='Slice padding', value=16, from_=4, to=24, increment=4)
        self.input_config.new_item(name='Scale image to (px)', value=16, from_=16,
                                   to=128, increment=16, on_change=self._enable_scaling)
        self.input_config_panel = gui.backend.ConfigPanel(master=self, config=self.input_config)
        self.input_config_panel.show(side=gui.interface.TOP, fill=gui.interface.X, expand=gui.interface.ON)
        self.training_config = datahandler.Config(panel_name='Training configuration:')
        self.training_config.new_item(name='Percent of data for validation', value=40,
                                      from_=10, to=50, increment=5)
        self.training_config.new_item(name='Batch size', value=8, from_=1, to=64)
        self.training_config.new_item(name='Epochs', value=75, from_=25, to=1000, increment=25)
        self.training_config.new_item(name='Number of networks to train', value=1, from_=1, to=50)
        # self.training_config.new_item(name='Keep best networks', value=3, from_=1, to=50)
        self.training_config_panel = gui.backend.ConfigPanel(master=self, config=self.training_config)
        self.training_config_panel.show(side=gui.interface.TOP, fill=gui.interface.X, expand=gui.interface.ON)
        self.network_config = datahandler.Config(panel_name='Network configuration:')
        self.network_config.new_item(name='Kernels', value=32, from_=4, to=128, increment=4)
        self.network_config.new_item(name='Conv3d layers', value=1, from_=1, to=5)
        self.network_config.new_item(name='ConvLSTM2D layers', value=4, from_=0, to=4)
        self.network_config.new_item(name='Reduction layers', value=3, from_=1, to=4)
        self.network_config_panel = gui.backend.ConfigPanel(master=self, config=self.network_config)
        self.network_config_panel.show(side=gui.interface.TOP, fill=gui.interface.X, expand=gui.interface.ON)
        self.figure_panel = gui.backend.Panel(master=self)
        self.model_factory = datahandler.models.SequenceClassificationModel(config=self.network_config)
        self.network_trainer = datahandler.models.NetworkTrainer(model_factory=self.model_factory)
        self.network_list_panel = gui.backend.NetworkListPanel(
            master=self.figure_panel, network_list=self.workflow_data.networks)
        self.network_list_panel.show(side=gui.interface.LEFT)
        self.figure_panel.show(side=gui.interface.TOP)
        self.buttons = gui.backend.Panel(master=self)
        self.done_button = gui.backend.Button(master=self.buttons, text='start training')
        self.done_button.show(side=gui.interface.LEFT)
        self.buttons.show(side=gui.interface.BOTTOM)
        self.manual_scale = False
        self.update()
        # TODO: (prio 3) reserve part of the training data for testing in the prediction tab
        # TODO: (prio 1) add option to filter out empty training stacks

    @property
    def input(self):
        '''Input data to be processed'''
        return (self.workflow_data.processed, self.workflow_data.annotation)

    @input.setter
    def input(self, inp: tuple):
        self.workflow_data.processed.update(inp[0])
        self.workflow_data.annotation.update(inp[1])

    @property
    def config(self):
        '''Configuration data'''
        return self._config

    @config.setter
    def config(self, conf: datahandler.interface.Config):
        self._config = conf

    @property
    def output(self):
        '''Output data after processing'''
        return self.workflow_data.networks

    def is_ready(self):
        '''Check if this step is ready to be run.'''
        return not self.workflow_data.processed.empty() and not self.workflow_data.annotation.empty()

    def is_done(self):
        '''Check if this step is done and ready for the next step.'''
        return self.is_ready() and not self.output.empty()

    def update(self):
        if self.is_ready():
            self.done_button.on_click = self.start_training
            if not self.manual_scale:
                self.input_config['Scale image to (px)'].value = next(
                    self.workflow_data.processed.image_stacks()).height

    def load_input(self):
        if self.workflow_data.path is None:
            fd = gui.backend.Filedialog()
            self.workflow_data.path = fd.ask_open(filetypes=([('numpy compressed', 'npz')]))
        (x_path, y_path) = self.default_in_file()
        y_path = self.file_versioning.get_newest_file(y_path)
        if os.path.exists(x_path) and y_path is not None:
            self.input[0].load(x_path)
            self.input[1].load(y_path)
        self.load_networks()
        # TODO: (prio 2) load/save config

    def save_output(self):
        self.output.save(self.default_out_file())

    def default_in_file(self):
        return (os.path.join(datahandler.get_eval_dir(self.workflow_data.path), 'junction_stacks.npz'), os.path.join(datahandler.get_eval_dir(self.workflow_data.path), 'junction_annotations.npz'))

    def default_out_file(self):
        return os.path.join(datahandler.get_eval_dir(self.workflow_data.path), 'network')

    def load_networks(self):
        if self.workflow_data.path is None:
            fd = gui.backend.Filedialog()
            self.workflow_data.path = fd.ask_save(filetypes=([]))
        self.workflow_data.networks.load(self.default_out_file())
        self.network_list_panel.update()

    def start_training(self):
        if self.manual_scale:
            dim = self.input_config['Scale image to (px)'].value
        else:
            dim = None
        self.slicer = datahandler.data.FormatTrainingData(
            slice_len=self.input_config['Slice length'].value,
            slice_padding=self.input_config['Slice padding'].value,
            target_dim=dim
        )
        self.data_generator = datahandler.models.TrainingDataGenerator(
            batch_size=self.training_config['Batch size'].value,
            len_sequences=self.input_config['Slice length'].value,
            test_size=self.training_config['Percent of data for validation'].value / 100)
        self.data_generator.set_data(*self.slicer.apply(*self.input))
        while self.data_generator.validation_steps() == 0 and self.training_config['Batch size'].value > 0:
            self.training_config['Batch size'].value -= 1
            self.data_generator.batch_size = self.training_config['Batch size'].value
        self.network_trainer.epochs = self.training_config['Epochs'].value
        self.network_trainer.training_data = self.data_generator
        self.network_trainer.num_categories = self.input[1].num_categories
        if self.fig_panel is None:
            self.fig_panel = gui.backend.FigurePanel(master=self.figure_panel, fig=self.network_trainer.get_fig())
            self.fig_panel.show(side=gui.interface.LEFT)
        self.network_trainer.set_canvas(self.fig_panel)
        if len(self.workflow_data.networks) == 0:
            self.load_networks()
        for _ in range(self.training_config['Number of networks to train'].value):
            trainer = self.network_trainer.copy()
            trainer.compile_model()
            trainer.train()
            unique_path = trainer.save(self.default_out_file())
            self.workflow_data.networks.append(trainer, unique_path)
            self.network_list_panel.update()

    def _enable_scaling(self, _=None):
        self.manual_scale = True



import bisect
import numpy as np
import matplotlib
import matplotlib.ticker
from matplotlib import pyplot as plt
from matplotlib.backend_bases import key_press_handler, button_press_handler
from .. import datahandler
from . import interface
_BACKEND = 'tkinter'
if _BACKEND == 'tkinter':
    from .tk import RootWindow, Panel, PanelKey, Filedialog, Dialogs, Button, Tabs, FigureCanvas, NetworkListPanel, ConfigEntry, ConfigScale, ConfigSpinbox, ConfigOptionSpinbox, ScaleSpinbox


class ConfigPanel(Panel, interface.ConfigPanel):
    '''create a panel element that contains a configuration GUI'''

    def __init__(self, master, config):
        Panel.__init__(self, master=master)
        if config.panel_name is not None:
            self.add_label(text=config.panel_name)
        self.item_list = []
        for item in config.values():
            self.add_item(item)

    def add_item(self, item: datahandler.interface.ConfigItem):
        if item.gui_element_type == 'entry':
            el = ConfigEntry(config_item=item, master=self)
        elif item.gui_element_type == 'slider':
            el = ConfigScale(config_item=item, master=self)
        elif item.gui_element_type == 'spinbox':
            if isinstance(item, datahandler.interface.OptionConfigItem):
                el = ConfigOptionSpinbox(config_item=item, master=self)
            else:
                el = ConfigSpinbox(config_item=item, master=self)
        else:
            raise NotImplementedError('Unrecognized gui element type: {}'.format(item.gui_element_type))
        self.item_list.append(el)
        el.show(side=interface.TOP, fill=interface.X, expand=interface.ON)


class FigurePanel(Panel, interface.FigurePanel):
    '''create a Panel that contains an interactive matplotlib.pyplot figure'''

    def __init__(self, master: Panel, fig=None, toolbar=True, width=None, height=None, aspect=None):
        Panel.__init__(self, master=master)
        self.aspect = aspect
        self.fig = fig or plt.figure()
        self._width = width or self.fig.get_size_inches()[0] * matplotlib.rcParams['figure.dpi']
        self._height = height or self.fig.get_size_inches()[1] * matplotlib.rcParams['figure.dpi']
        self.change_aspect(aspect, width is None)
        self.figure = FigureCanvas(fig, master=self)
        self.mouse = None
        self.toolbar = None
        self.figure.draw()
        if toolbar:
            self.add_toolbar()
        self.figure.show_canvas()

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, width):
        self._width = width or self.height * self.aspect
        self._update_fig_size()

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, height):
        self._height = height or self.width / self.aspect
        self._update_fig_size()

    def update(self):
        self.figure.draw()

    def add_mouse(self, index, press_callback=None, drag_callback=None, release_callback=None):
        self.mouse = FigurePanelMouse(self, press_callback=press_callback,
                                      drag_callback=drag_callback, release_callback=release_callback)

    def add_toolbar(self):
        self.toolbar = self.figure.get_toolbar(self)

    def remove_toolbar(self):
        self.toolbar.destroy()
        self.toolbar = None

    def add_save_btn(self):
        self.save_btn = self.figure.get_save_button(self)

    def remove_save_btn(self):
        self.save_btn.destroy()
        self.save_btn = None

    def destroy(self):
        plt.close(self.fig)
        super().destroy()

    def change_aspect(self, aspect, adjust_width=True):
        if aspect is not None:
            if adjust_width:
                self.width = self.height * aspect
            else:
                self.height = self.width / aspect
        self.aspect = self.width / self.height
        self._update_fig_size()

    def _update_fig_size(self):
        self.fig.set_size_inches(
            self.width / matplotlib.rcParams['figure.dpi'],
            self.height / matplotlib.rcParams['figure.dpi'])


class FigurePanelMouse:
    '''handles mouse events for FigurePanel'''

    def __init__(self, figure_canvas, press_callback=interface._empty_callback, drag_callback=interface._empty_callback, release_callback=interface._empty_callback, scroll_callback=interface._empty_callback):
        self.figure_canvas = figure_canvas
        self.press_callback = press_callback
        self.drag_callback = drag_callback
        self.release_callback = release_callback
        self.scroll_callback = scroll_callback
        self.scroll_id = self.figure_canvas.figure.mpl_connect('scroll_event', self.scroll_callback)
        self.mouse_press_id = self.figure_canvas.figure.mpl_connect("button_press_event", self.on_button_press)

    def on_button_press(self, event):
        if event.button == 1:
            self.mouse_drag_id = self.figure_canvas.figure.mpl_connect(
                'motion_notify_event',
                self.on_drag)
            self.mouse_release_id = self.figure_canvas.figure.mpl_connect(
                'button_release_event',
                self.on_release)
        self.press_callback(event)
        button_press_handler(event, self.figure_canvas.figure, self.figure_canvas.toolbar)

    def on_drag(self, event):
        self.drag_callback(event)

    def on_release(self, event):
        self.figure_canvas.figure.mpl_disconnect(self.mouse_drag_id)
        self.figure_canvas.figure.mpl_disconnect(self.mouse_release_id)
        self.release_callback(event)

    def destroy(self):
        self.figure_canvas.figure.mpl_disconnect(self.mouse_press_id)
        self.figure_canvas.figure.mpl_disconnect(self.scroll_id)


class FigurePanelWatcher:
    def __init__(self, figure_canvas, enter_callback=interface._empty_callback, leave_callback=interface._empty_callback):
        self.figure_canvas = figure_canvas
        self.enter_callback = enter_callback
        self.leave_callback = leave_callback
        self.enter_id = self.figure_canvas.figure.mpl_connect('figure_enter_event', self.enter_callback)
        self.leave_id = self.figure_canvas.figure.mpl_connect('figure_leave_event', self.leave_callback)

    def destroy(self):
        self.figure_canvas.figure.mpl_disconnect(self.enter_id)
        self.figure_canvas.figure.mpl_disconnect(self.leave_id)


class FigurePanelKey:
    '''handles keyboard events for FigurePanel'''

    def __init__(self, figure_canvas, callback=None):
        self.figure_canvas = figure_canvas
        self.callback = callback or interface._empty_callback
        self.key_press_id = self.figure_canvas.figure.mpl_connect("key_press_event", self.on_key_press)

    def on_key_press(self, event):
        self.callback(event.key)
        key_press_handler(event, self.figure_canvas.figure, self.figure_canvas.toolbar)

    def destroy(self):
        self.figure_canvas.figure.mpl_disconnect(self.key_press_id)


class ImagePanel(FigurePanel, interface.ImagePanel):
    '''Panel object that contains an interactive matplotlib.pyplot figure'''

    def __init__(self, master: Panel, image: np.ndarray, toolbar=True, zoom=1, cmap='grays'):
        '''initialize ImagePanel
        Arguments:
        master: Panel into which figure will be placed
        image: numpy ndarray
        toolbar: bool enable or disable the pyplot navigation toolbar (optional, default true)
        '''
        if isinstance(zoom, tuple):
            self.im = datahandler.rotatable.Image(image=image, figsize=zoom, cmap=cmap)
        else:
            self.im = datahandler.rotatable.ZoomImage(image=image, zoom=zoom, cmap=cmap)
        FigurePanel.__init__(self, master=master, fig=self.im.fig, toolbar=toolbar)

    def update(self, image=None):
        '''refresh the figure
        Arguments:
        image: np.ndarray replaces the image data (optional)
        '''
        self.im.update(image)
        self.figure.draw()

    def get_ax(self):
        return self.im.ax


class RotatableImagePanel(ImagePanel, interface.RotatableImagePanel):

    def __init__(self, master, image, toolbar=True, zoom=1):
        self.im = datahandler.rotatable.RotatableFigure(image, zoom=zoom)
        FigurePanel.__init__(self, master=master, fig=self.im.fig, toolbar=toolbar)

    @property
    def rot(self):
        '''integer: stores the rotation in radians'''
        return self.im.rot

    @rot.setter
    def rot(self, rot: int):
        self.im.rot = rot

    @property
    def flipped(self):
        '''boolean: stores the flipped state'''
        return self.im.flipped

    @flipped.setter
    def flipped(self, flipped: bool):
        self.im.flipped = flipped

    def add_controls(self, master=None):
        if master is None:
            master = self
        self.image_controls = Panel(master=master)
        self.rotate_button = Button(
            self.image_controls,
            text='rotate',
            on_click=self.rotate)
        self.rotate_button.show(side=interface.LEFT)
        self.flip_button = Button(
            self.image_controls,
            text='flip',
            on_click=self.flip)
        self.flip_button.show(side=interface.LEFT)
        self.image_controls.show(side=interface.BOTTOM, fill=interface.BOTH, expand=interface.ON)

    def rotate(self):
        self.im.rotate()
        self.update()
        try:
            self.rotation_callback()
        except AttributeError:
            pass

    def flip(self):
        self.im.flip()
        self.update()
        try:
            self.flip_callback()
        except AttributeError:
            pass


class LinkedImages(Panel, interface.ImagePanel):
    def __init__(self, master, image_list, toolbar=False, zoom=1):
        Panel.__init__(self, master)
        self.image_list = image_list
        self.canvas_list = []
        self.mouse_list = []
        for i, image in enumerate(self.image_list):
            self.mouse_list.append(None)
            self.canvas_list.append(RotatableImagePanel(master=self, image=np.squeeze(image),
                                                        toolbar=toolbar, zoom=zoom))
            if i > 0:
                self.canvas_list[i].rotation_callback = self.canvas_list[0].rotate
                self.canvas_list[i].flip_callback = self.canvas_list[0].flip
            self.canvas_list[i].show(side=interface.LEFT, fill=interface.X, expand=interface.ON, padx=2, pady=2)

    def update(self, image_list=None):
        if image_list is not None:
            self.image_list = image_list
        for i in range(len(self.canvas_list)):
            self.canvas_list[i].update(self.image_list[i])

    def add_mouse(self, index, press_callback=None, drag_callback=None, release_callback=None):
        '''add mouse event callback handler to the matplotlib figure
        Arguments:
        index:  int index of the image that the callbacks should be assigned to.
                If index is not an int, the callback will be assigned to all figures.
        press_callback (optional): function handle executed on mouse button press
        drag_callback (optional): function handle executed on mouse drag (with button pressed)
        release_callback (optional): function handle executed on mouse button release
        '''
        if type(index, int):
            self.mouse_list[index] = FigurePanelMouse(self.canvas_list[index],
                                                      press_callback=press_callback,
                                                      drag_callback=drag_callback,
                                                      release_callback=release_callback)
        else:
            for n, canvas in enumerate(self.canvas_list):
                self.mouse_list[n] = FigurePanelMouse(canvas,
                                                      press_callback=press_callback,
                                                      drag_callback=drag_callback,
                                                      release_callback=release_callback)

    def add_controls(self):
        self.canvas_list[0].add_controls(self)

    def get_ax(self, index):
        return self.canvas_list[index].get_ax()


class LinkedStacks(Panel, interface.StackPanel):
    def __init__(self, master, stack_list, toolbar=False, zoom=1):
        Panel.__init__(self, master)
        self.controls = StackControls(master=self, callback=self.update_frame, key_ud=False)
        self.on_frame_change = interface._empty_callback
        self._stacks = stack_list
        self.controls.max_frame = self.max_frame
        self.images = LinkedImages(self, self.image_list,
                                   toolbar=toolbar, zoom=zoom)
        self.controls.show(side=interface.BOTTOM, fill=interface.X, expand=interface.ON)
        self.images.show(side=interface.BOTTOM)

    @property
    def stack_list(self):
        return self._stacks

    @stack_list.setter
    def stack_list(self, stack_list):
        self._stacks = stack_list
        self.controls.max_frame = self.max_frame
        self.update()

    @property
    def frame(self):
        return self.controls.frame

    @frame.setter
    def frame(self, frame):
        self.controls.frame = frame
        self.update()
        self.on_frame_change(frame)

    @property
    def frames_of_interest(self):
        return self.controls.frames_of_interest

    @frames_of_interest.setter
    def frames_of_interest(self, foi: list):
        self.controls.frames_of_interest = foi

    @property
    def on_frame_change(self):
        return self._on_frame_change

    @on_frame_change.setter
    def on_frame_change(self, on_frame_change):
        self._on_frame_change = on_frame_change

    @property
    def max_frame(self):
        return min([stack.shape[0] for stack in self.stack_list]) - 1

    @property
    def image_list(self):
        return [np.squeeze(stack[self.frame]) for stack in self.stack_list]

    def add_mouse(self, *args, **kw):
        self.images.add_mouse(*args, **kw)

    def update(self):
        self.images.update(self.image_list)

    def get_ax(self, index):
        return self.images.get_ax(index)

    def update_frame(self, frame=None):
        if frame is not None:
            self.frame = frame
        else:
            self.update()
            self.on_frame_change(self.frame)


class MultiStacks(Panel):
    def __init__(self, master, multi_stack: datahandler.interface.MultiImageStack, toolbar=False, zoom=1):
        Panel.__init__(self, master)
        self._multi_stack = None
        self.frames_of_interest = list()
        self.controls = StackControls(master=self, callback=self.update_stack, key_lr=False)
        self.multi_stack = multi_stack
        self.stack_display = LinkedStacks(master=self,
                                          stack_list=[self.multi_stack[0]],
                                          toolbar=toolbar,
                                          zoom=zoom)
        self.stack_display.add_label(text=self.multi_stack.names[0])
        self.controls.add_label(text='switch stack:')
        self.stack_display.show(side=interface.TOP)
        self.controls.show(side=interface.TOP, fill=interface.X, expand=interface.ON)
        self.on_frame_change = interface._empty_callback
        self.on_stack_change = interface._empty_callback

    @property
    def multi_stack(self):
        return self._multi_stack

    @multi_stack.setter
    def multi_stack(self, ms: datahandler.interface.MultiImageStack):
        self._multi_stack = ms
        self.controls.max_frame = len(self._multi_stack) - 1

    @property
    def stack_name(self):
        return self._multi_stack.names[self.controls.frame]

    @property
    def frame(self):
        return self.stack_display.frame

    @frame.setter
    def frame(self, frame):
        self.update_frame(frame=frame)

    @property
    def on_frame_change(self):
        return self.stack_display.on_frame_change

    @on_frame_change.setter
    def on_frame_change(self, on_frame_change):
        self.stack_display.on_frame_change = on_frame_change

    def update(self, multi_stack=None):
        if isinstance(multi_stack, datahandler.interface.MultiImageStack):
            self.multi_stack = multi_stack
        self.update_stack()

    def update_stack(self, i=None):
        if i is None:
            i = self.controls.frame
        self.frame = 0
        self.stack_display.label['text'] = self.multi_stack.names[i]
        self.stack_display.stack_list = [self.multi_stack[i]]
        if len(self.frames_of_interest) == len(self.multi_stack):
            self.stack_display.frames_of_interest = self.frames_of_interest[i]
        self.on_stack_change(i)

    def update_frame(self, frame=None):
        self.stack_display.update_frame(frame=frame)

    def next_stack(self, inc=1):
        self.controls.frame += inc


class ImageStackClassification(Panel):
    def __init__(self, master, multi_stack: datahandler.interface.MultiImageStackClassification, toolbar=False, zoomed_height=385):
        Panel.__init__(self, master)
        self.fast_seeker = FastSeeker()
        self.height = 7
        self.threshold = 0.5
        self.changed = False
        self.frames_of_interest = list()
        self.multi_stack = multi_stack
        self._stack = 0
        self._frame = 0
        self.zoomed_height = zoomed_height
        self.toolbar = toolbar
        self.image_panel = None
        self.mouse = None
        self.watcher = None
        self.frame_indicator = None
        self.on_scroll = interface._empty_callback
        self.figure_leave_callback = interface._empty_callback
        self.over_scroll = 0
        self.under_scroll = 0
        self.over_scroll_threshold = 12
        self.on_over_scroll = interface._empty_callback
        self.create_figure()

    @property
    def multi_stack(self):
        return self._multi_stack

    @multi_stack.setter
    def multi_stack(self, ms: datahandler.interface.MultiImageStackClassification):
        self._multi_stack = ms
        self.changed = False

    @property
    def frame(self):
        return self._frame

    @frame.setter
    def frame(self, frame):
        self._frame = int(frame)

    @property
    def stack(self):
        return self._stack

    @stack.setter
    def stack(self, stack):
        self._stack = int(stack)

    @property
    def offset(self):
        return int(np.floor(self.height / 2))

    @property
    def total_frames(self):
        return self.multi_stack[self.stack].shape[0]

    @property
    def max_frame(self):
        return self.total_frames - 1

    def create_figure(self):
        area = self.current_area()
        shape = area.shape
        self.image_panel = ImagePanel(self, area, cmap='magma',
                                      toolbar=self.toolbar, zoom=self.zoomed_height/self.height)
        self.image_panel.show()
        self.mouse = FigurePanelMouse(self.image_panel, press_callback=self.toggle, scroll_callback=self.scroll_frame)
        self.watcher = FigurePanelWatcher(self.image_panel, leave_callback=self.on_figure_leave)
        self.frame_indicator = datahandler.rotatable.RotatableRectangle(
            pos=(-0.5, -0.5), size=(shape[1], 1), image_shape=shape)
        self.update_indicator()
        self.frame_indicator.draw(self.image_panel.get_ax())

    def change_frame(self, frame):
        if frame < 0:
            frame = 0
            self.under_scroll += 1
        elif frame > 0:
            self.under_scroll = 0
        if frame > self.max_frame:
            frame = self.max_frame
            self.over_scroll += 1
        elif frame < self.max_frame:
            self.over_scroll = 0
        self.frame = frame
        self.update()
        if self.over_scroll > self.over_scroll_threshold:
            self.over_scroll = 0
            self.on_over_scroll(1)
        elif self.under_scroll > self.over_scroll_threshold:
            self.under_scroll = 0
            self.on_over_scroll(-1)

    def change_stack(self, i):
        self.stack = i
        self.update()

    def current_area(self):
        data = self.multi_stack[self.stack]
        if self.frame < self.offset:
            return data[:self.height, :]
        elif self.frame < data.shape[0] - self.offset:
            return data[self.frame - self.offset:self.frame + self.offset + 1, :]
        else:
            return data[-self.height:, :]

    def re_draw(self):
        self.image_panel.destroy()
        self.create_figure()

    def update(self):
        self.update_indicator()
        self.image_panel.update(self.current_area())
        if len(self.frames_of_interest) == len(self.multi_stack):
            self.fast_seeker.frames_of_interest = self.frames_of_interest[self.stack]

    def update_indicator(self):
        if self.frame < self.offset:
            self.frame_indicator.pos = (-0.5, self.frame - 0.5)
        elif self.frame < self.total_frames - self.offset:
            self.frame_indicator.pos = (-0.5, self.offset - 0.5)
        else:
            self.frame_indicator.pos = (-0.5, self.height + self.frame - self.total_frames - 0.5)
        self.frame_indicator.update()

    def toggle(self, event):
        x = int(round(event.xdata))
        y = int(round(event.ydata))
        if self.frame < self.offset:
            pass
        elif self.frame < self.total_frames - self.offset:
            y += self.frame - self.offset
        else:
            y += self.total_frames - self.height
        self.multi_stack[self.stack][y, x] = float(self.multi_stack[self.stack][y, x] < self.threshold)
        self.changed = True
        self.update()

    def scroll_frame(self, event):
        if event.key == 'shift':
            if event.step < 0:
                self.change_frame(self.fast_seeker.fast_forward(self.frame, self.max_frame))
            if event.step > 0:
                self.change_frame(self.fast_seeker.fast_rewind(self.frame, self.max_frame))
        elif event.key == 'control':
            self.change_frame(self.frame - 2 * event.step)
        else:
            self.change_frame(self.frame - event.step)
        self.on_scroll(self.frame)

    def add_toolbar(self):
        self.image_panel.add_toolbar()

    def on_figure_leave(self, event=None):
        self.figure_leave_callback()


class MultiStacksAnnotations(Panel):
    def __init__(self,
                 master,
                 multi_stack: datahandler.interface.MultiImageStack,
                 annotation=datahandler.MultiImageStackClassification(),
                 toolbar=False,
                 display_height=384,
                 copy_predictions=interface._empty_callback):
        Panel.__init__(self, master)
        self.frames_of_interest = None
        self.foi_config = datahandler.Config(panel_name=' ')
        self.foi_config.new_item(name='Frame of interest threshold', value=40, from_=0.5,
                                 to=100, increment=0.5, on_change=self._calc_foi)
        self.stacks = multi_stack
        self.display_height = display_height
        height = self.stacks.image_stacks(0).height
        self.stack_gui = MultiStacks(master=self, multi_stack=self.stacks, zoom=self.display_height/height)
        self.stack_gui.show(side=interface.LEFT)
        self.annotation = annotation
        self.annotation.create_from_multi_image_stack(self.stacks)
        self.annotation_panel = Panel(master=self)
        self.annotation_panel_heading = Panel(master=self.annotation_panel)
        self.annotation_panel_labels = []
        self.refresh_annotation_labels()
        self.annotation_panel_heading.show(side=interface.TOP, fill=interface.X)
        self.annotation_gui = ImageStackClassification(
            master=self.annotation_panel, multi_stack=self.annotation, zoomed_height=self.display_height)
        self.stack_gui.on_stack_change = self.annotation_gui.change_stack
        self.stack_gui.on_frame_change = self.annotation_gui.change_frame
        self.annotation_gui.on_scroll = self.stack_gui.update_frame
        self.annotation_gui.on_over_scroll = self.stack_gui.next_stack
        self.annotation_gui.show(side=interface.TOP, padx=2, pady=2)
        self.annotation_config = Panel(self.annotation_panel)
        self.annotation_config.show(side=interface.LEFT)
        self.annotation_panel.show(side=interface.LEFT, fill=interface.Y, expand=interface.ON)
        self.dialogs = Dialogs(title='Annotation')
        self.output_config = datahandler.Config(panel_name='Output configuration:')
        self.output_config.new_item(name='Number of Paths', gui_element_type='spinbox', value=6,
                                    from_=2, to=10, on_change=self.numpaths_update)
        self.copy_predictions = copy_predictions
        self.foi_config_panel = ConfigPanel(master=self.annotation_config, config=self.foi_config)
        self.foi_config_panel.show(side=interface.BOTTOM)
        self._calc_foi()

    @property
    def stacks(self):
        return self._stacks

    @stacks.setter
    def stacks(self, ms: datahandler.interface.MultiImageStack):
        self._stacks = ms
        self._calc_foi()

    @property
    def num_categories(self):
        return self.output_config['Number of Paths'].value

    @num_categories.setter
    def num_categories(self, cat):
        self.output_config['Number of Paths'].value = cat

    @property
    def on_annotation_leave(self):
        return self.annotation_gui.figure_leave_callback

    @on_annotation_leave.setter
    def on_annotation_leave(self, callback):
        self.annotation_gui.figure_leave_callback = callback

    @property
    def annotation_changed(self):
        return self.annotation_gui.changed

    @annotation_changed.setter
    def annotation_changed(self, val: bool):
        self.annotation_gui.changed = val

    def add_config_ui(self):
        self.output_config_panel = ConfigPanel(self.annotation_config, self.output_config)
        self.output_config_panel.show(side=interface.TOP)

    def update(self, multi_stack=None):
        if isinstance(multi_stack, datahandler.interface.MultiImageStack):
            self.stacks = multi_stack
            self.annotation.update_multi_stack(self.stacks)
            self.annotation_gui.change_stack(0)
            self.stack_gui.update(self.stacks)
        else:
            self.stack_gui.update()
        self.annotation_gui.update()

    def refresh_annotation_labels(self):
        for i, name in enumerate(self.annotation.category_names):
            try:
                self.annotation_panel_labels[i].edit_label(text=name)
            except IndexError:
                self.annotation_panel_labels.append(Panel(master=self.annotation_panel_heading))
                self.annotation_panel_labels[-1].add_label(text=name)
            self.annotation_panel_labels[i].show(side=interface.LEFT, fill=interface.X, expand=interface.ON)
        for n in range(i + 1, len(self.annotation_panel_labels)):
            self.annotation_panel_labels[n].hide()

    def numpaths_update(self, num_paths):
        '''Make sure the number of columns of output matches num_paths'''
        try:
            self.annotation.num_categories = num_paths
        except datahandler.interface.VidtrainDataLossException:
            if self.dialogs.yes_no('Reducing the number of categories will cause data loss. Are you sure?'):
                self.annotation.force = True
                self.annotation.num_categories = num_paths
            else:
                self.output_config['Number of Paths'].value = self.annotation.num_categories
        except AttributeError:
            pass
        try:
            self.annotation_gui.multi_stack = self.annotation
            self.refresh_annotation_labels()
            self.annotation_gui.re_draw()
        except AttributeError:
            pass

    def clear_annotations(self):
        self.annotation.create_from_multi_image_stack(self.stacks)
        self.annotation_gui.update()

    def add_toolbar(self):
        self.annotation_gui.add_toolbar()

    def _calc_foi(self, _=None):
        self.frames_of_interest = list()
        for stack in self.stacks:
            iqr = np.percentile(stack, 0.75) - np.percentile(stack, 0.25)
            med = np.percentile(stack, 0.5)
            threshold = med + iqr * self.foi_config['Frame of interest threshold'].value
            self.frames_of_interest.append(np.where(np.any(stack > threshold, axis=(1, 2, 3)))[0])
        try:
            self.stack_gui.frames_of_interest = self.frames_of_interest
            self.stack_gui.update()
            self.annotation_gui.frames_of_interest = self.frames_of_interest
            self.annotation_gui.update()
        except AttributeError:
            pass


class ClassificationEvalPlot(FigurePanel):
    def __init__(self, master, isc=None, plot='classes', **kw):
        FigurePanel.__init__(self, master, **kw)
        self.classification = isc
        self.axes = self.fig.add_subplot(1, 1, 1)
        self.plot = plot
        self.title = ''

    def update(self, isc: datahandler.interface.ImageStackClassification):
        self.classification = isc
        self.axes.clear()
        self.num_cat = self.classification.data.shape[1]
        self.bar_x = list(self.classification.category_names)[:self.num_cat]
        if self.plot == 'classes':
            self.plot_classes()
        elif self.plot == 'combined':
            self.plot_combined()
        else:
            self.plot_frames()
        self.axes.set_title(self.title)
        self.fig.tight_layout()
        super().update()

    def plot_classes(self):
        self.axes.bar(self.bar_x, np.sum(self.classification.data, axis=0))
        self.axes.set_xlabel('class')
        self.axes.set_ylabel('events')
        self.axes.yaxis.set_major_locator(matplotlib.ticker.MaxNLocator(integer=True, min_n_ticks=1))
        self.axes.tick_params(axis='x', labelrotation=90)

    def plot_frames(self):
        frames = np.array(range(self.classification.data.shape[0]))
        for cat in range(self.num_cat):
            self.axes.plot(frames, self.classification.data[:, cat], label=self.bar_x[cat])
        self.axes.set_xlabel('frame')
        self.axes.set_ylabel('probability')
        self.axes.legend()
        self.axes.yaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter('%.1f'))
        self.axes.xaxis.set_major_locator(matplotlib.ticker.MaxNLocator(integer=True, prune='both'))
        self.axes.set_ylim(0, 1)

    def plot_combined(self):
        junction_names = self.classification._category_names
        self.axes.bar(junction_names, self.classification.data[:, 0])
        self.axes.set_xlabel('junction_coordinates')
        self.axes.set_ylabel('events')
        self.axes.yaxis.set_major_locator(matplotlib.ticker.MaxNLocator(integer=True, prune='both'))
        self.axes.tick_params(axis='x', labelrotation=90)


class ClassificationEvaluationPlots(Panel):
    def __init__(self, master, isc=None, height=384):
        Panel.__init__(self, master)
        self.per_class = ClassificationEvalPlot(self, fig=plt.figure(), isc=isc, height=height, aspect=0.5)
        self.per_time = ClassificationEvalPlot(self, fig=plt.figure(), isc=isc,
                                               plot='frames', height=height, aspect=2)
        self.per_class.show()
        self.per_time.show()

    def update(self, isc: datahandler.interface.ImageStackClassification):
        self.per_class.update(isc)
        self.per_time.update(isc)


class StackControls(Panel):
    def __init__(self, master, callback, key_lr=True, key_ud=True):
        Panel.__init__(self, master)
        self.fast_seeker = FastSeeker()
        self.callback = callback
        self.frame_slider = ScaleSpinbox(master=self, to=0, on_change=self.on_change)
        self.controls = Panel(master=self)
        self.skip_start_btn = Button(master=self.controls, text='|<', on_click=self.skip_start)
        self.ffwd_btn = Button(master=self.controls, text='>>', on_click=self.fast_forward)
        # TODO: (prio 4) add play-forward, stop, play-backward, buttons
        self.frwd_btn = Button(master=self.controls, text='<<', on_click=self.fast_rewind)
        self.skip_end_btn = Button(master=self.controls, text='>|', on_click=self.skip_end)
        self.skip_start_btn.pack(side=interface.LEFT)
        self.frwd_btn.pack(side=interface.LEFT)
        self.ffwd_btn.pack(side=interface.LEFT)
        self.skip_end_btn.pack(side=interface.LEFT)
        self.controls.pack(side=interface.BOTTOM)
        self.frame_slider.pack(side=interface.BOTTOM, fill=interface.BOTH, expand=1)
        self.key_lr = key_lr
        self.key_ud = key_ud
        if key_lr or key_ud:
            self.panel_key = PanelKey(master=master.winfo_toplevel(), callback=self.on_key_pressed)

    @property
    def frame(self):
        return self.frame_slider.value

    @frame.setter
    def frame(self, val):
        if val > self.max_frame:
            val = self.max_frame
        if val < 0:
            val = 0
        if self.frame_slider.value != int(val):
            self.frame_slider.value = int(val)
            self.on_change()

    @property
    def max_frame(self):
        return self.frame_slider.to

    @max_frame.setter
    def max_frame(self, limit):
        self.frame_slider.to = limit

    @property
    def frames_of_interest(self):
        return self.fast_seeker.frames_of_interest

    @frames_of_interest.setter
    def frames_of_interest(self, foi: list):
        self.fast_seeker.frames_of_interest = foi

    def skip_start(self):
        self.frame = 0

    def skip_end(self):
        self.frame = self.max_frame

    def fast_forward(self):
        self.frame = self.fast_seeker.fast_forward(self.frame, self.max_frame)

    def fast_rewind(self):
        self.frame = self.fast_seeker.fast_rewind(self.frame, self.max_frame)

    def on_change(self, frame=None):
        self.callback(self.frame)

    def on_key_pressed(self, key):
        if (key.lower() == 'right' and self.key_lr) or (key.lower() == 'up' and self.key_ud):
            self.frame += 1
        elif (key.lower() == 'left' and self.key_lr) or (key.lower() == 'down' and self.key_ud):
            self.frame -= 1
        elif (key.lower() == 'shift+right' and self.key_lr) or (key.lower() == 'shift+up' and self.key_ud):
            self.fast_forward()
        elif (key.lower() == 'shift+left' and self.key_lr) or (key.lower() == 'shift+down' and self.key_ud):
            self.fast_rewind()
        elif (key.lower() == 'control+right' and self.key_lr) or (key.lower() == 'control+up' and self.key_ud):
            self.frame += 2
        elif (key.lower() == 'control+left' and self.key_lr) or (key.lower() == 'control+down' and self.key_ud):
            self.frame -= 2


class FastSeeker:
    def __init__(self, frames_of_interest=[]):
        self.frames_of_interest = frames_of_interest

    def fast_forward(self, frame, max_frame):
        if len(self.frames_of_interest) > 0:
            try:
                frame = self.frames_of_interest[bisect.bisect_right(self.frames_of_interest, frame)]
            except IndexError:
                frame = self.frames_of_interest[-1]
        else:
            frame += round(max_frame / 10) + 1
        return frame

    def fast_rewind(self, frame, max_frame):
        if len(self.frames_of_interest) > 0:
            index = bisect.bisect_left(self.frames_of_interest, frame)
            if frame in self.frames_of_interest and index > 0:
                index -= 1
            try:
                frame = self.frames_of_interest[index]
            except IndexError:
                frame = self.frames_of_interest[-1]
        else:
            frame -= round(max_frame / 10) + 1
        return frame



import re
import copy
import os
import pathlib
import glob
import collections
import warnings
from time import strftime
import numpy as np
import pandas as pd
import micdata
from . import interface
from . import config
from . import models


def get_eval_dir(path):
    (dirname, basename) = os.path.split(path)
    (name, _) = os.path.splitext(basename)
    if dirname[-5:] == '_eval':
        evaldir = dirname
    else:
        evaldir = os.path.join(dirname, name + '_eval')
    pathlib.Path(evaldir).mkdir(parents=True, exist_ok=True)
    return evaldir


class ImageStack(interface.ImageStack):
    def __init__(self, data=None, normalizer=micdata.formatting.StackNormalizer()):
        self.normalizer = normalizer
        self.data = data
        self.flatten_formatter = micdata.formatting.ParallelFlattener(function=np.median, kw={'axis': 0})

    @property
    def data(self):
        '''the raw data'''
        return self._data

    @data.setter
    def data(self, data):
        if data is not None:
            assert isinstance(data, np.ndarray)
            if data.ndim == 2:
                data = data[np.newaxis, :, :, np.newaxis]
            elif data.ndim == 3:
                data = data[..., np.newaxis]
            assert data.ndim == 4
            data = self.normalizer.apply(data)
        self._data = data

    @property
    def num_slices(self):
        '''return: int: number of images in the stack'''
        return self.data.shape[0]

    @property
    def width(self):
        '''return: int: width of each image in pixels'''
        return self.data.shape[1]

    @property
    def height(self):
        '''return: int: height of each image in pixels'''
        return self.data.shape[2]

    @property
    def image_shape(self):
        '''return: tuple: shape of each image in pixels'''
        return self.data.shape[1:]

    @property
    def num_channels(self):
        '''return: int: number of color channels of each image'''
        return self.data.shape[3]

    def load(self, path):
        self.data = micdata.io.IOFactory(path).load()

    def save(self, path):
        if self.data is not None:
            micdata.io.IOFactory(path).write(self.data, ext='.tif')

    def median(self):
        '''return: 2D np.ndarray: median projection of the stack'''
        self.flatten_formatter.parameters['function'] = np.median
        return np.squeeze(self.flatten_formatter.apply(self.data))

    def mean(self):
        '''return: 2D np.ndarray: mean projection of the stack'''
        self.flatten_formatter.parameters['function'] = np.mean
        return np.squeeze(self.flatten_formatter.apply(self.data))

    def std(self):
        '''return: 2D np.ndarray: standard deviation projection of the stack'''
        self.flatten_formatter.parameters['function'] = np.std
        return np.squeeze(self.flatten_formatter.apply(self.data))

    def max(self):
        '''return: 2D np.ndarray: maximum projection of the stack'''
        return np.squeeze(np.max(self.data, axis=0))

    def copy_tile(self, pos, dim):
        '''copy a tile out of the stack.
        Arguments:
        tuple pos: x and y coordinates
        typle dim: wdth and height of the tile
        return: 4D np.ndarray: copy a part of the stack or None if the tile does not fit onto the stack'''
        tile = np.copy(self.data[:, pos[1]: pos[1] + dim[1], pos[0]: pos[0] + dim[0], :])
        if tile.shape[1:3] < dim:
            raise IndexError
        return tile

    def empty(self):
        return self.data is None or self.data.size == 0


class MultiImageStack(interface.MultiImageStack):
    '''many image stacks stored as collections.MutableSequence'''

    def __init__(self, *args, positions=[]):
        self.data = list(args)
        if len(positions) == len(self.data):
            self._parse_all_pos(positions)
        else:
            self.positions = [None] * len(self.data)

    @property
    def names(self):
        return ['x:{:04d} y:{:04d}'.format(pos[0], pos[1]) if isinstance(pos, tuple) else str(i) for i, pos in enumerate(self.positions)]

    def __len__(self): return len(self.data)

    def __getitem__(self, i):
        return self.data[i]

    def __setitem__(self, i, v: tuple):
        self._check(v)
        self.data[i] = v[0]
        self.positions[i] = self._parse_pos(v[1])

    def __delitem__(self, i):
        del self.data[i]
        del self.positions[i]

    def __iter__(self):
        yield from self.data

    def __add__(self, other: interface.MultiImageStack):
        self.data += other.data
        self.positions += other.positions
        return self

    def insert(self, i, v: tuple):
        self._check(v)
        self.data.insert(i, v[0])
        self.positions.insert(i, self._parse_pos(v[1]))

    def get_position(self, i):
        return self.positions[i]

    def append(self, stack, position):
        self.data.append(stack)
        self.positions.append(position)

    def np_array(self):
        '''return an ndarray from data.values. only works if the shape of all values is the same'''
        return np.array(self.data)

    def image_stacks(self, i=None):
        '''return a generator that yields ImageStacks.
        if i is provided, an individual ImageStack is created from the data corresponding to that index.
        '''
        if i is None:
            return (ImageStack(data=v) for v in self.data)
        return ImageStack(data=self.data[i])

    def update_from(self, template: interface.MultiImageStack):
        self.data = template.data
        self.positions = template.positions

    def load(self, path):
        assert path[-4:] == '.npz', 'File must be a .npz file. Got {}'.format(path)
        data = dict(np.load(path, allow_pickle=False))
        try:
            with open(self._positions_file(path), 'r') as f:
                self._parse_all_pos(f.read().splitlines())
        except FileNotFoundError:
            self._parse_all_pos(data.keys())
        self.data = list(data.values())

    def save(self, path):
        np.savez_compressed(path, *self.data)
        with open(self._positions_file(path), 'w') as f:
            for pos in self.positions:
                f.write(str(pos) + '\n')

    def empty(self):
        return len(self.data) == 0

    def copy(self):
        return copy.deepcopy(self)

    def from_dict(self, d: dict):
        self.data = list(d.values())
        self._parse_all_pos(d.keys())

    def _parse_pos(self, pos):
        if isinstance(pos, tuple):
            return pos
        x = y = False
        if pos[0] == '(':
            x = re.search(r'\((\d+),', pos)
            y = re.search(r',\s*(\d+)\)', pos)
        if pos[0] == 'x' or pos[0] == 'y':
            x = re.search(r'x:(\d+)', pos)
            y = re.search(r'y:(\d+)', pos)
        if x and y:
            return (int(x.group(1)), int(y.group(1)))
        return None

    def _parse_all_pos(self, pos_list):
        self.positions = [self._parse_pos(p) for p in pos_list]

    def _positions_file(self, path):
        return path + '.positions'

    def _check(self, v):
        assert len(v) == 2, 'value must be a tuple of length 2: (data: np.ndarray, position: tuple or None). Got {}.'.format(str(v))
        assert isinstance(v[0], np.ndarray)


class ImageStackClassification(interface.ImageStackClassification):
    def __init__(self, data=None, category_names=None):
        self.data = data
        self.category_names = category_names or []
        self.force = False

    @property
    def data(self):
        '''the raw data'''
        return self._data

    @data.setter
    def data(self, data: np.ndarray):
        if data is not None:
            assert isinstance(data, np.ndarray)
            assert data.ndim == 2
        self._data = data

    @property
    def category_names(self):
        n = 0
        while n < self.num_categories:
            try:
                yield self._category_names[n]
            except IndexError:
                yield 'cat {}'.format(n)
            n += 1

    @category_names.setter
    def category_names(self, names):
        self._category_names = names

    @property
    def num_slices(self):
        '''return: int: number of images in the stack'''
        return self.data.shape[0]

    @property
    def num_categories(self):
        '''the number of categories'''
        return self.data.shape[1]

    @num_categories.setter
    def num_categories(self, categories):
        if self.num_categories < categories:
            self.data = np.pad(self.data, ((0, 0), (0, categories - self.num_categories)))
        elif self.num_categories > categories:
            if np.any(self.data[:, categories:] != 0) and not self.force:
                raise interface.VidtrainDataLossException(
                    'Reducing the number of categories will lead to data loss. Add force=True to override this error.')
            self.data = self.data[:, : categories]
            self.force = False

    def load(self, path):
        assert path[-4:] == '.npz', 'File must be a .npz file. Got {}'.format(path)
        with np.load(path, allow_pickle=False) as loaded:
            self.data = loaded['arr_0']

    def save(self, path):
        np.savez_compressed(path, self.data)

    def empty(self):
        return len(self.data) == 0


class MultiImageStackClassification(MultiImageStack, interface.MultiImageStackClassification):
    def __init__(self, multi_image_stack=None, num_categories=6, all_category_names=['A1', 'A2', 'B1', 'B2', 'detach', 'land', 'AB', 'BA', 'AA', 'BB', '1A', '1B', '2A', '2B', '12', '21', '11', '22']):
        self.data = []
        self._num_categories = num_categories
        self.all_category_names = all_category_names
        self.force = False
        if multi_image_stack is not None:
            self.create_from_multi_image_stack(multi_image_stack)

    @property
    def category_names(self):
        n = 0
        while n < self.num_categories:
            try:
                yield self.all_category_names[n]
            except IndexError:
                yield 'cat {}'.format(n)
            n += 1

    @category_names.setter
    def category_names(self, names):
        self.all_category_names = names

    @property
    def num_categories(self):
        '''the number of categories'''
        return self._num_categories

    @num_categories.setter
    def num_categories(self, categories):
        if self._num_categories != categories:
            self._test_data_loss(categories)
            self._num_categories = categories
            for i, data in enumerate(self.data):
                if data.shape[1] < categories:
                    self.data[i] = np.pad(data, ((0, 0), (0, categories - data.shape[1])))
                elif data.shape[1] > categories:
                    if np.any(self.data[i][:, categories:] != 0) and not self.force:
                        raise interface.VidtrainDataLossException(
                            'Reducing the number of categories will lead to data loss. Set the force property to True to override this error.')
                    self.data[i] = data[:, :categories]
            self.force = False

    def save(self, path):
        super().save(path)
        with open(self._category_file_name(path), 'w') as f:
            f.write('\n'.join(self.all_category_names))

    def load(self, path):
        super().load(path)
        self._num_categories = self.data[0].shape[1]
        try:
            with open(self._category_file_name(path), 'r') as f:
                self.all_category_names = f.readlines()
        except FileNotFoundError:
            pass

    def update_from(self, template: interface.MultiImageStackClassification):
        self.data = copy.deepcopy(template.data)
        self.positions = template.positions.copy()
        self.num_categories = template.num_categories
        self.all_category_names = template.all_category_names

    def update_multi_stack(self, mis: interface.MultiImageStack):
        temp = self.data.copy()
        self.create_from_multi_image_stack(mis)
        if len(temp) >= len(self.data):
            self.data = temp[0:len(self.data)]
        else:
            self.data[0:len(temp)] = temp

    def image_stacks(self, key=None):
        '''return a generator that yields ImageStacks.
        if key is provided, an individual ImageStack is created from the data corresponding to that key.
        '''
        if key is None:
            return (ImageStackClassification(data=v, category_names=self.all_category_names) for v in self.data)
        return ImageStackClassification(data=self.data[key], category_names=self.all_category_names)

    def create_from_multi_image_stack(self, mis: interface.MultiImageStack):
        self.data = []
        for stack in mis:
            self.data.append(np.zeros((stack.shape[0], self.num_categories)))
        self.positions = mis.positions

    def apply_threshold(self, threshold):
        for i, val in enumerate(self.data):
            self.data[i] = (val > threshold).astype(np.float32)

    def _test_data_loss(self, categories):
        for d in self.data:
            if d.shape[1] > categories:
                if np.any(d[:, categories:] != 0) and not self.force:
                    raise interface.VidtrainDataLossException(
                        'Reducing the number of categories will lead to data loss. Add force=True to override this error.')

    def _category_file_name(self, path):
        return path + '.categories'


class FormatTrainingData(interface.FormatTrainingData):
    def __init__(self, slice_len, slice_padding, target_dim=None):
        self.slice_len = slice_len
        self.slice_padding = slice_padding
        self.target_dim = target_dim
        self.reversible = True
        self.data_structure = []

    def apply(self, x: interface.MultiImageStack, y: interface.MultiImageStackClassification):
        '''slice the numpy arrays in dict x and y into slices and concatenate the slices into a list
        x and y must have identical keys and the first dimension of the numpy arrays must match
        '''
        x_out = []
        y_out = []
        slicer = micdata.formatting.StackSlicer(slice_len=self.slice_len, slice_padding=self.slice_padding)
        if self.target_dim is not None:
            scaler = micdata.formatting.StackScaler(target_dim=self.target_dim)
        for i, _ in enumerate(x):
            if self.target_dim is not None:
                self.reversible = False
                x[i] = scaler.apply(x[i])
            if x[i].shape[0] > self.slice_len:
                x_list = list(slicer.apply(x[i]))
                self.data_structure.append((i, len(x_list)))
                x_out += x_list
                y_out += list(slicer.apply(y[i][0]))
            else:
                self.reversible = False
                x_out += [x[i]]
                y_out += [y[i]]
        return (x_out, y_out)


class FileVersioning():
    TIMEFORMAT = '%Y-%m-%d_%H%M'
    TIMEFORMAT_REGEX = r'_\d{4}-\d{2}-\d{2}_\d{4}'

    def __init__(self, base_path=None):
        self.base_path = base_path

    def get_name(self, base_path=None, unique=True):
        base_path = base_path or self.base_path
        if unique:
            (path_prefix, ext) = os.path.splitext(base_path)
            return path_prefix + '_' + strftime(self.TIMEFORMAT) + ext
        else:
            return base_path

    def get_file_list(self, base_path=None):
        base_path = base_path or self.base_path
        (path_prefix, ext) = os.path.splitext(base_path)
        return glob.glob(path_prefix + '*' + ext)

    def get_newest_file(self, base_path=None):
        base_path = base_path or self.base_path
        try:
            return self.get_file_list(base_path)[-1]
        except IndexError:
            return None

    def get_part_before_time(self, base_path=None):
        base_path = base_path or self.base_path
        return re.split(self.TIMEFORMAT_REGEX, base_path, maxsplit=1)[0]


class SequencePredictionFormatter(interface.PredictionFormatter):
    def __init__(self, sequence_len=64, time_overlap=16):
        self.sequence_len = sequence_len
        self.time_overlap = time_overlap
        self.stack_sizes = []
        self.unique_frames = self.sequence_len - self.time_overlap
        assert self.unique_frames > 0, 'Time overlap must be smaller than sequence length. Got sequence_len: {} and time_overlap: {}'.format(
            self.sequence_len, self.time_overlap)
        self.formatter = micdata.formatting.StackSlicer(slice_len=self.unique_frames, slice_padding=self.time_overlap)

    def apply(self, data: interface.MultiImageStack):
        stacks = []
        for stack in data:
            self.stack_sizes.append(stack.shape[0])
            stacks.append(stack)
        stack = np.vstack(stacks)
        return self.formatter.apply(stack)

    def revert(self, data: interface.MultiImageStack, y_pred):
        padding = sum(self.stack_sizes) % self.unique_frames
        reverted = self.formatter.revert(y_pred)
        if padding > 0:
            remove = self.unique_frames - padding
            reverted = reverted[:-remove]
        result = MultiImageStackClassification(multi_image_stack=data, num_categories=y_pred.shape[-1])
        split_at = np.cumsum(self.stack_sizes[:-1])
        for i, stack in enumerate(np.array_split(reverted, split_at)):
            result.data[i] = stack
        return result


class NetworkList(interface.NetworkList):
    def __init__(self):
        self.networks = pd.DataFrame({'Median loss': [], 'Network object': [], 'Network filename': []})
        self.file_versioning = FileVersioning()

    def append(self, network, file_name):
        self.networks.loc[len(self.networks)] = [self._calc_loss(network), network, file_name]
        self.networks.sort_values(by=['Median loss'], inplace=True)  # add ignore_index=True for pandas version 1.0.0
        self.networks.reset_index(drop=True, inplace=True)

    def update_from(self, template: interface.NetworkList):
        self.networks = template.networks

    def load(self, path_prefix):
        for f in self.file_versioning.get_file_list(path_prefix + '.h5'):
            self.load_network(f.rpartition('_')[0])

    def save(self, path_prefix):
        for network in self.networks.loc[:, 'Network object']:
            network.save(path_prefix)

    def empty(self):
        return len(self.networks) == 0

    def load_network(self, file_name_no_ext):
        if file_name_no_ext not in list(self.networks.loc[:, 'Network filename']):
            network = models.NetworkTrainer()
            network.load(file_name_no_ext)
            self.append(network, file_name_no_ext)

    def get_network(self, index):
        return self.networks.loc[index, 'Network object']

    def clear(self):
        self.networks.drop(list(self.networks.index), inplace=True)
        print(self.networks)  # needed to avoid race condition

    def predict(self, data: interface.MultiImageStack, formatter: interface.PredictionFormatter, best=None):
        if best is None:
            best = self.networks.shape[0]
        split_data = formatter.apply(data)
        num_categories = self.networks.loc[0, 'Network object'].num_categories
        result = np.zeros((best,) + split_data.shape[:2] + (num_categories,))
        for n, network in self.networks['Network object'].items():
            if n == best:
                break
            result[n] = network.model.predict(split_data)
        median = formatter.revert(data, np.median(result, axis=0))
        std = formatter.revert(data, np.std(result, axis=0))
        return (median, std)

    @staticmethod
    def _calc_loss(network):
        return np.median(network.history.history['val_loss'][-10:])

    def __len__(self):
        return len(self.networks)


class JunctionAnalysisData(interface.WorkflowData):
    def __init__(self,
                 path=None,
                 raw=ImageStack(),
                 processed=MultiImageStack(),
                 annotation=MultiImageStackClassification(),
                 networks=NetworkList(),
                 pred_med=MultiImageStackClassification(),
                 pred_std=MultiImageStackClassification(),
                 pred_corr=MultiImageStackClassification()):
        self.raw = raw
        self.processed = processed
        self.annotation = annotation
        self.networks = networks
        self.pred_med = pred_med
        self.pred_std = pred_std
        self.pred_corr = pred_corr
        self.path = path
        # TODO: (prio 3) move save and load methods from workflow steps here

    @property
    def path(self):
        '''path where the data will be stored'''
        return self._path

    @path.setter
    def path(self, path):
        self._path = path



import os
import numpy as np
from .. import gui
from .. import datahandler
from . import interface
# TODO: (prio 5) enable storing and loading training data and trained networks in a central database


class JunctionAnnotation(gui.backend.Panel, interface.WorkflowStep):
    NAME = 'Annotate Junction Crossing Events'

    def __init__(self, master, workflow_data: datahandler.interface.WorkflowData):
        gui.backend.Panel.__init__(self, master=master)
        self.workflow_data = workflow_data
        self.gui = None
        self.copy_prediction_button = None
        self.prediction_threshold_config = None
        self.autosave = False
        self.file_versioning = datahandler.FileVersioning()
        if self.is_ready():
            self.update()
        # TODO: (prio 2) enable adding training data from another source

    @property
    def input(self):
        '''Input data to be annotated'''
        return self.workflow_data.processed

    @input.setter
    def input(self, inp: datahandler.interface.MultiImageStack):
        self.self.workflow_data.processed.update(inp)
        self.update()

    @property
    def config(self):
        '''Configuration data'''
        return self._config

    @config.setter
    def config(self, conf: datahandler.interface.Config):
        self._config = conf

    @property
    def output(self):
        '''Output data after processing'''
        return self.workflow_data.annotation

    def update(self):
        if self.is_ready():
            try:
                self.gui.update(self.input)
            except AttributeError:
                self.create_gui()
            if self.autosave and self.workflow_data.path is not None:
                self.gui.on_annotation_leave = self.auto_save_output
            else:
                self.gui.on_annotation_leave = gui.interface._empty_callback
            if not self.workflow_data.pred_corr.empty():
                self.add_copy_predictions_ui()

    def create_gui(self):
        if self.is_ready():
            self.gui = gui.backend.MultiStacksAnnotations(master=self, multi_stack=self.input, annotation=self.output)
            self.gui.add_config_ui()
            self.gui.show(side=gui.interface.LEFT)
        # TODO: (prio 3) add configuration for class names (use listbox and entry widgets)

    def is_ready(self):
        '''Check if this step is ready to be run.'''
        return not self.input.empty()

    def is_done(self):
        '''Check if this step is done and ready for the next step.'''
        try:
            return self.is_ready and not self.output.empty()
        except AttributeError:
            return False

    def load_input(self):
        if self.workflow_data.path is None:
            fd = gui.backend.Filedialog()
            self.workflow_data.path = fd.ask_open(filetypes=([('numpy compressed', 'npz')]))
        if os.path.exists(self.default_in_file()):
            self.input.load(self.default_in_file())
            self.update()
        try:
            self.output.load(self.file_versioning.get_newest_file(self.default_out_file()))
        except TypeError:
            pass

    def save_output(self, unique=True):
        assert self.is_done, 'Cannot save output data without input. Please set/load input data first.'
        if self.workflow_data.path is None:
            fd = gui.backend.Filedialog()
            self.workflow_data.path = fd.ask_save(filetypes=([('numpy compressed', 'npz')]))
        out_path = self.file_versioning.get_name(base_path=self.default_out_file(), unique=unique)
        self.output.save(out_path)

    def auto_save_output(self):
        if self.gui.annotation_changed:
            self.save_output()
            self.gui.annotation_changed = False

    def default_in_file(self):
        return os.path.join(datahandler.get_eval_dir(self.workflow_data.path), 'junction_stacks.npz')

    def default_out_file(self):
        return os.path.join(datahandler.get_eval_dir(self.workflow_data.path), 'junction_annotations.npz')

    def add_copy_predictions_ui(self):
        if self.copy_prediction_button is None:
            self.copy_prediction_button = gui.backend.Button(
                self.gui.annotation_config, text='copy predictions', on_click=self.copy_predictions)
            self.copy_prediction_button.show(side=gui.interface.BOTTOM)
            self.prediction_threshold_config = datahandler.Config()
            self.prediction_threshold_config.new_item(name='Prediction trheshold', value=0.5,
                                                      from_=0.1, to=0.9, on_change=self.copy_predictions)
            self.prediction_threshold_config_panel = gui.backend.ConfigPanel(
                self.gui.annotation_config, self.prediction_threshold_config)
            self.prediction_threshold_config_panel.show(side=gui.interface.BOTTOM)

    def copy_predictions(self, _=None):
        self.workflow_data.annotation.update_from(self.workflow_data.pred_corr)
        self.workflow_data.annotation.apply_threshold(self.prediction_threshold_config['Prediction trheshold'].value)
        self.gui.update()



import abc
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from ..datahandler import interface as data_interface

# Booleans
NO = FALSE = OFF = 0
YES = TRUE = ON = 1

# -fill
NONE = 'none'
X = 'x'
Y = 'y'
BOTH = 'both'

# -side
LEFT = 'left'
TOP = 'top'
RIGHT = 'right'
BOTTOM = 'bottom'

# -orient
HORIZONTAL = 'horizontal'
VERTICAL = 'vertical'


def _empty_callback(event=None):
    pass


class RootWindow(metaclass=abc.ABCMeta):
    '''Create an application window'''

    @abc.abstractmethod
    def run(self):
        '''Run the application'''


class Element(metaclass=abc.ABCMeta):
    '''any gui element needs to specify a master'''
    @abc.abstractmethod
    def __init__(self, master):
        '''initialize the panel and define its master container'''

    @abc.abstractmethod
    def show(self, side=LEFT, fill=NONE, expand=OFF, padx=0, pady=0):
        '''show the element'''

    @abc.abstractmethod
    def hide(self):
        '''hide the element'''

    @abc.abstractmethod
    def destroy(self):
        '''destroy the element'''


class Panel(Element):
    '''A user interface container for other elements'''

    @abc.abstractmethod
    def add_label(self, text, side=TOP, fill=NONE, expand=OFF):
        '''add a label to the panel'''

    @abc.abstractmethod
    def edit_label(self, text):
        '''edit the label text'''


class FigurePanel(Panel):
    '''Panel object that contains an interactive matplotlib.pyplot figure'''
    @abc.abstractmethod
    def __init__(self, master: Panel, fig: plt.Figure, toolbar=True, **kw):
        '''initialize FigurePanel
        Arguments: 
        master: Panel into which figure will be placed
        fig: pyplot Figure object 
        toolbar (optional): bool enable or disable the pyplot navigation toolbar (default true)
        '''

    @abc.abstractmethod
    def update(self):
        '''refresh the figure'''

    @abc.abstractmethod
    def add_mouse(self, press_callback=None, drag_callback=None, release_callback=None):
        '''add mouse event callback handler to the matplotlib figure
        Arguments:
        press_callback (optional): function handle executed on mouse button press
        drag_callback (optional): function handle executed on mouse drag (with button pressed)
        release_callback (optional): function handle executed on mouse button release
        '''

    @abc.abstractmethod
    def destroy(self):
        '''close figure and call super().destroy()'''


class ImagePanel(FigurePanel):
    '''Panel object that contains an interactive matplotlib.pyplot figure'''

    @abc.abstractmethod
    def __init__(self, master: Panel, image: np.ndarray, toolbar=True, **kw):
        '''initialize ImagePanel
        Arguments: 
        master: Panel into which figure will be placed
        image: numpy ndarray 
        toolbar (optional): bool enable or disable the pyplot navigation toolbar (default true)
        '''

    @abc.abstractmethod
    def update(self, image=None):
        '''refresh the figure
        Arguments:
        image (optional): np.ndarray replaces the image data 
        '''


class RotatableImagePanel(ImagePanel, data_interface.Rotatable):
    '''StackPanel that also implements data_interface.Rotatable'''

    @abc.abstractmethod
    def add_controls(self, master=None):
        '''add rotation controls
        Arguments:
        master: Panel to which the controls should be added
        '''


class StackPanel(ImagePanel):
    '''StackPanel object that contains controls to show several frames of an image stack'''

    @property
    def frame(self):
        '''int: current frame'''

    @property
    def on_frame_change(self):
        '''callable: callback that is executed once the frame changes'''


class ConfigElement(metaclass=abc.ABCMeta):
    pass


class ConfigPanel(Panel):
    '''create a panel element that contains a configuration GUI'''

    @abc.abstractmethod
    def add_item(self, item: data_interface.ConfigItem):
        '''Add a config item'''


class Filedialog(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def ask_open(self, filetypes):
        '''Open a filedialog and return the path to a file.

        Arguments:
        filetypes: 2d tuple of strings describing the allowed file types 
                   (('description', '*.ext'))

        Returns:
        string: path to file
        '''

    @abc.abstractmethod
    def ask_save(self, filetypes):
        '''Open a filedialog and return the path to a file.

        Arguments:
        filetypes: 2d tuple of strings describing the allowed file types 
                   (('description', '*.ext'))

        Returns:
        string: path to file
        '''

    @abc.abstractmethod
    def ask_directory(self):
        '''open a directory'''


class Dialogs(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __init__(self, title=None):
        '''set the title for dialogs during initialization'''

    @abc.abstractmethod
    def info(self, message):
        '''Opens a info message dialog with an ok button'''

    @abc.abstractmethod
    def yes_no(self, message):
        '''Opens a yes/no message dialog'''


class Button(Element):
    '''an interactive button'''

    @property
    @abc.abstractmethod
    def on_click(self):
        '''callback function that is executed when the button is clicked'''



from numbers import Number
import abc
import warnings
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from . import interface


class RotatableMixin(interface.Rotatable):
    '''Mixin class for rotatable objects'''

    def __init__(self, rot=0, flipped=False):
        self.rot = rot
        self.flipped = flipped

    @property
    def rot(self):
        return self.__rot

    @rot.setter
    def rot(self, rot):
        self.__rot = rot % 4

    @property
    def flipped(self):
        return self.__flipped

    @flipped.setter
    def flipped(self, flipped):
        self.__flipped = flipped

    def rotate(self):
        self.rot += 1

    def flip(self):
        self.flipped = not self.flipped


class RotatableRectangle(RotatableMixin):
    def __init__(self, pos, size, image_shape, flipped=False, rot=0, edgecolor='w'):
        assert len(image_shape) == 2
        assert isinstance(flipped, bool)
        assert isinstance(rot, int)
        self.image_shape = image_shape
        self.flipped = flipped
        self.rot = rot
        self.dim = size
        self.pos = pos
        self.rect = patches.Rectangle(
            self.pos,
            self.dim[0], self.dim[1],
            linewidth=1,
            edgecolor=edgecolor,
            facecolor='none')

    @property
    def dim(self):
        if (self.rot % 2) > 0:
            return (self.__dim[1], self.__dim[0])
        else:
            return self.__dim

    @dim.setter
    def dim(self, size):
        try:
            if len(size) == 2:
                if (self.rot % 2) > 0:
                    self.__dim = (size[1], size[0])
                else:
                    self.__dim = tuple(size)
            else:
                raise ValueError('Rectangle dimensions must be one or two numbers. Got: {}'.format(size))
        except TypeError:
            self.__dim = (size, size)
        try:
            self.rect.set_width(self.dim[0])
            self.rect.set_height(self.dim[1])
        except AttributeError:
            pass

    @property
    def pos(self):
        return self._rotate_coords(self.__pos)

    @pos.setter
    def pos(self, pos):
        if isinstance(pos, str) and pos == 'center':
            pos = np.round((np.asarray(self.image_shape) - np.asarray(self.dim)) / 2)
        self.__pos = self._unrotate_coords(pos)

    @property
    def raw_pos(self):
        return self.__pos

    @raw_pos.setter
    def raw_pos(self, pos):
        self.__pos = tuple(pos)

    def update(self):
        self.rect.set_xy(self.pos)

    def update_from(self, rect):
        if isinstance(rect, RotatableRectangle):
            self.image_shape = rect.image_shape
            self.flipped = rect.flipped
            self.rot = rect.rot
            self.dim = rect.dim
            self.pos = np.copy(rect.pos)
            rect = rect.rect
        self.rect.update_from(rect)
        self.update()

    def draw(self, ax):
        ax.add_patch(self.rect)

    def remove(self):
        try:
            self.rect.remove()
        except NotImplementedError:
            pass

    def extract_image(self, rotatable_image):
        pos = list(map(int, self.pos))
        dim = list(map(int, self.dim))
        try:
            try:
                rotated = rotatable_image.rotated_image()
            except AttributeError:
                rotated = rotatable_image
                if self.rot > 0 or self.flipped:
                    warnings.warn(
                        'Could not rotate image. Use a RotatableImage object or make sure you rotated the input yourself.')
            result = np.copy(rotated[pos[1]: pos[1] + dim[1], pos[0]: pos[0] + dim[0]])
        except IndexError:
            result = np.zeros(dim)
        if np.any(np.asarray(result.shape) == 0):
            result = np.zeros(dim)
        return result

    def _rotate_coords(self, coord):
        if self.flipped:
            coord = self._flip_coords(coord)
        return self._value_rotation(coord, self.rot)

    def _unrotate_coords(self, coord):
        coord = self._value_rotation(coord, 4 - self.rot)
        if self.flipped:
            coord = self._flip_coords(coord)
        return coord

    def _flip_coords(self, coord):
        return (self.image_shape[0] - coord[0] - self.dim[0], coord[1])

    def _value_rotation(self, coord, rotation):
        if rotation == 1:
            coord = (coord[1], self.image_shape[1] - coord[0] - self.dim[1])
        elif rotation == 2:
            coord = (self.image_shape[0] - coord[0] - self.dim[0],
                     self.image_shape[1] - coord[1] - self.dim[1])
        elif rotation == 3:
            coord = (self.image_shape[0] - coord[1] - self.dim[0], coord[0])
        elif rotation > 4:
            raise ValueError('Rotation must be 4 or less. Got {}'.format(rotation))
        return tuple(coord)


class RectangleList(RotatableMixin):
    def __init__(self):
        self.rect_list = []

    @property
    def pos(self):
        try:
            return self.rect_list[0].pos
        except IndexError:
            return None

    @pos.setter
    def pos(self, pos):
        try:
            offset = map(lambda a, b: a - b, pos, self.rect_list[0].pos)
            for rect in self.rect_list:
                rect.pos = map(lambda a, b: a + b, offset, rect.pos)
        except IndexError:
            pass

    @property
    def dim(self):
        try:
            return self.rect_list[0].dim
        except IndexError:
            return None

    @dim.setter
    def dim(self, size):
        for rect in self.rect_list:
            rect.dim = size

    @property
    def rot(self):
        try:
            return self.rect_list[0].rot
        except IndexError:
            return None

    @rot.setter
    def rot(self, rot):
        for rect in self.rect_list:
            rect.rot = rot

    @property
    def flipped(self):
        try:
            return self.rect_list[0].flipped
        except IndexError:
            return None

    @flipped.setter
    def flipped(self, flipped):
        for rect in self.rect_list:
            rect.flipped = flipped

    def flip(self):
        for rect in self.rect_list:
            rect.flip()

    def rotate(self):
        for rect in self.rect_list:
            rect.rotate()

    def update(self):
        for rect in self.rect_list:
            rect.update()

    def add_rect(self, pos, edgecolor='w', size=None, image_shape=None, rot=None, flipped=None):
        if size is None:
            try:
                size = self.rect_list[0].dim
            except IndexError:
                size = 16
        if image_shape is None:
            try:
                image_shape = self.rect_list[0].image_shape
            except IndexError:
                image_shape = (512, 512)
        if rot is None:
            try:
                rot = self.rect_list[0].rot
            except IndexError:
                rot = 0
        if flipped is None:
            try:
                flipped = self.rect_list[0].flipped
            except IndexError:
                flipped = False
        self.rect_list.append(RotatableRectangle(pos, size, image_shape, flipped, rot, edgecolor))
        # TODO (prio 5) use rect_list[-1].update_from(rect_list[0]) and possibly switch to IdenticalList

    def delete_rect(self, index=-1):
        self.rect_list.pop(index).remove()

    def remove_all(self):
        for rect in self.rect_list:
            rect.remove()

    def add_all(self, ax):
        for rect in self.rect_list:
            ax.add_patch(rect.rect)

    def refresh_ax(self, ax):
        self.remove_all()
        self.add_all(ax)


class RectangleMatrix(RectangleList):
    def __init__(self, image_shape=(512, 512), num_rows=1, num_cols=16, row_height=16, rot=0, flipped=False, size=16):
        super().__init__()
        from numbers import Number
        assert isinstance(num_rows, int)
        assert isinstance(num_cols, int)
        assert isinstance(row_height, Number)
        self.num_rows = num_rows
        self.num_cols = num_cols
        self._row_height = row_height
        left_pos = [0, image_shape[1]]
        self.add_rect(left_pos, edgecolor='r', size=size, image_shape=image_shape, rot=rot, flipped=flipped)
        self.left_rect = self.rect_list[0]
        right_pos = [image_shape[0], image_shape[1]]
        self.add_rect(right_pos, edgecolor='b', size=size, image_shape=image_shape, rot=rot, flipped=flipped)
        self.right_rect = self.rect_list[1]
        self.current_rect = self.left_rect
        self.update_rect_number()

    @property
    def row_height(self):
        return self._row_height

    @row_height.setter
    def row_height(self, height):
        self._row_height = height
        self.distribute()

    def generator(self):
        indices = list(range(len(self.rect_list)))
        indices.append(1)
        del(indices[1])  # we need to put the second element last so that the rectangles are in the order they appear on screen
        for i in indices:
            yield self.rect_list[i]

    def update_rect_number(self):
        target = self.num_rows * self.num_cols
        diff = target - len(self.rect_list)
        if diff > 0:
            for _ in range(diff):
                self.add_rect(self.left_rect.pos)
        elif diff < 0:
            for _ in range(abs(diff)):
                self.delete_rect()
        if diff != 0:
            self.distribute()

    def distribute(self):
        cur_rect = 2
        x_dist = self.right_rect.pos[0] - self.left_rect.pos[0]
        y_dist = self.right_rect.pos[1] - self.left_rect.pos[1]
        for row in range(self.num_rows):
            x_col_step = (x_dist) / (self.num_cols - 1)
            y_col_step = (y_dist) / (self.num_cols - 1)
            if abs(x_dist) < abs(y_dist):
                row_x = self.left_rect.pos[0] - row * self.row_height
                row_y = self.left_rect.pos[1]
                if y_col_step != 0:
                    row_y += row * x_col_step * self.row_height / y_col_step
            else:
                row_x = self.left_rect.pos[0]
                row_y = self.left_rect.pos[1] - row * self.row_height
                if x_col_step != 0:
                    row_x += row * y_col_step * self.row_height / x_col_step
            start = 1 if row == 0 else 0
            for col in range(start, self.num_cols - start):
                pos = [row_x + col * x_col_step, row_y + col * y_col_step]
                self.rect_list[cur_rect].pos = pos
                cur_rect += 1
        self.update()

    def store_current_pos(self, pos):
        self.current_rect.pos = pos
        self.distribute()

    def move_current_rect(self, key):
        update = False
        pos = list(self.current_rect.pos)
        if str(key).lower() in ['up', '8']:
            pos[1] -= 1
            update = True
        elif str(key).lower() in ['down', '2']:
            pos[1] += 1
            update = True
        elif str(key).lower() in ['left', '4']:
            pos[0] -= 1
            update = True
        elif str(key).lower() in ['right', '6']:
            pos[0] += 1
            update = True
        if update:
            self.current_rect.pos = pos
            self.distribute()

    def select_left(self):
        self.current_rect = self.left_rect

    def select_right(self):
        self.current_rect = self.right_rect

    def toggle_corner(self):
        if self.current_rect == self.right_rect:
            self.select_left()
        else:
            self.select_right()


class Image(interface.Image):
    '''create a matplotlib image with fixed dimensions'''

    def __init__(self, image: np.ndarray, figsize: tuple, cmap='gray'):
        self._fig = plt.figure(figsize=figsize)
        self.im_obj = plt.imshow(image, cmap=cmap, animated=True, vmin=0, vmax=1)
        self._ax = self.im_obj.axes
        self.ax.axis('off')
        self.ax.set_position([0, 0, 1, 1])

    @property
    def fig(self):
        '''matplotlib figure'''
        return self._fig

    @property
    def ax(self):
        '''matplotlib axes'''
        return self._ax

    def update(self, image: np.ndarray):
        '''update the image with new data
        Arguments:
        image: numpy ndarray
        '''
        self.image_shape = image.shape
        self.im_obj.set_array(image)
        # self.fig.set_size_inches(*self._figsize(), forward=True)

    def __del__(self):
        plt.close(self.fig)


class ZoomImage(Image):
    '''create a matplotlib image that can easily be zoomed'''

    def __init__(self, image: np.ndarray, zoom=1, cmap='gray'):
        '''Arguments:
        image: numpy ndarray
        zoom: the zoom factor (optional)
        '''
        assert isinstance(zoom, Number)
        self.image_shape = image.shape
        self.zoom = zoom
        Image.__init__(self, image=image, figsize=self._figsize(), cmap=cmap)

    def _figsize(self):
        return (self.image_shape[1] / 72 * self.zoom, self.image_shape[0] / 72 * self.zoom)


class RotatableFigure(RotatableMixin, interface.ZoomImage):
    def __init__(self, image, flipped=False, rot=0, zoom=1):
        assert isinstance(flipped, bool)
        assert isinstance(rot, int)
        self.image = np.asarray(image)
        self.flipped = flipped
        self.rot = rot
        self.zoom_fig = ZoomImage(self.rotated_image(), zoom=zoom)

    @property
    def fig(self):
        return self.zoom_fig.fig

    @property
    def ax(self):
        return self.zoom_fig.ax

    def rotated_image(self):
        rotated = np.copy(self.image)
        if self.flipped:
            rotated = np.fliplr(rotated)
        if self.rot > 0:
            rotated = np.rot90(rotated, self.rot)
        return rotated

    def rotate_stack(self, stack):
        if self.flipped:
            stack = stack[:, :, ::-1, ...]
        if self.rot > 0:
            stack = np.rot90(stack, self.rot, axes=(1, 2))
        return stack

    def update(self, image=None):
        if image is not None:
            self.image = image
        self.zoom_fig.update(self.rotated_image())

    def __del__(self):
        del(self.zoom_fig)



import pickle
import gzip
import collections
from . import interface


def create_config_item(name, value, **kw):
    '''Factory method for config items.'''
    if type(value) == str:
        return StringConfigItem(name=name, value=value, **kw)
    elif type(value) == int:
        if 'options' in kw.keys():
            return OptionConfigItem(name=name, value=value, **kw)
        else:
            return IntConfigItem(name=name, value=value, **kw)
    elif type(value) == float:
        return FloatConfigItem(name=name, value=value, **kw)
    elif type(value) == dict:
        return Config(panel_name=name, **value)
    elif isinstance(value(interface.Config, interface.ConfigItem)):
        return value
    else:
        raise ValueError('Value must be str, int, float, ConfigItem, dict or Config. Got {}'.format(type(value)))


def _empty_callback(val=None):
    pass


class Config(interface.Config):
    def __init__(self, *args, panel_name=None, **kw):
        self.panel_name = panel_name
        collections.UserDict.__init__(self, *args, **kw)

    def new_item(self, name, value, **kw):
        self.data[name] = create_config_item(name, value, **kw)

    def update(self, d):
        for name, value in d.items():
            if isinstance(value, dict):
                self.data[name] = create_config_item(name, **value)
            else:
                self.data[name] = create_config_item(name, value)

    def save(self, path):
        with gzip.open(path, 'w') as f:
            pickle.dump(self.data, f)

    def load(self, path):
        with gzip.open(path, 'r') as f:
            loaded = pickle.load(f)
        assert isinstance(loaded, dict), 'file must contain a dict. Got" {}'.format(type(loaded))
        self.data = loaded


class ConfigItemMixin:
    DEFAULT_GUI_EL = 'entry'

    def __init__(self, name: str, value, gui_element_type='entry', on_change=_empty_callback):
        self.name = name
        self._value = self.convert_val(value)
        self.default_value = value
        self.gui_element_type = gui_element_type
        self.on_change = on_change

    @property
    def value(self):
        '''Any: Value of the configuration item'''
        return self._value

    @value.setter
    def value(self, val):
        val = self.convert_val(val)
        change = val != self._value
        self._value = val
        if change:
            self.on_change(val)

    @property
    def default_value(self):
        '''Any: Default value of the configuration item'''
        return self._default_value

    @default_value.setter
    def default_value(self, val):
        self._default_value = self.convert_val(val)

    @property
    def gui_element_type(self):
        '''The gui element that should be used by the user to set the config item'''
        return self._gui_element_type

    @gui_element_type.setter
    def gui_element_type(self, element_type):
        if element_type not in self.GUI_ELEMENT_TYPES:
            element_type = self.DEFAULT_GUI_EL
        self._gui_element_type = element_type

    @property
    def on_change(self):
        '''Callable: A callback that is executed when the value changes'''
        return self._on_change

    @on_change.setter
    def on_change(self, callback):
        self._on_change = callback

    def is_valid(self, val):
        try:
            test = self.convert_val(val)
            return test is not None
        except Exception:
            return False

    def reset(self):
        '''Reset value to default_value.'''
        self.value = self.default_value

    def name2var(self) -> str:
        '''convert the name to a variable name'''
        return self.name.lower().replace(' ', '_')


class StringConfigItem(ConfigItemMixin, interface.ConfigItem):

    def convert_val(self, val):
        '''Ensure that val is a string (convert it if necessary).'''
        return str(val)


class OptionConfigItem(ConfigItemMixin, interface.OptionConfigItem):
    '''Configuration item that allows selecting one or more predefined options'''
    GUI_ELEMENT_TYPES = ['spinbox', 'radio', 'checkbox']
    DEFAULT_GUI_EL = 'spinbox'

    def __init__(self, name, value, options, gui_element_type='spinbox', on_change=_empty_callback):
        self.on_options_change = _empty_callback
        self.options = options
        ConfigItemMixin.__init__(self, name=name, value=value,
                                 gui_element_type=gui_element_type, on_change=on_change)

    @property
    def options(self):
        '''list of options'''
        return self._options

    @options.setter
    def options(self, option_list):
        assert isinstance(option_list, list)
        self._options = option_list
        if hasattr(self, '_value'):
            self.value = self.value  # ensure that value is still valid
        self.on_options_change()

    @property
    def current_option(self):
        return self.options[self.value]

    @current_option.setter
    def current_option(self, option):
        self.value = self.options.index(option)

    @property
    def to(self):
        return len(self.options) - 1

    @property
    def from_(self):
        return 0

    @property
    def on_options_change(self):
        return self._on_options_change

    @on_options_change.setter
    def on_options_change(self, callback):
        self._on_options_change = callback

    def toggle(self):
        if self.value == self.to:
            self.value = 0
        else:
            self.value += 1

    def convert_val(self, val):
        '''Ensure that val is a valid index of options.'''
        val = int(val)
        if val > self.to:
            val = self.to
        if val < self.from_:
            val = self.from_
        return val


class FloatConfigItem(ConfigItemMixin, interface.NumericConfigItem):
    DEFAULT_GUI_EL = 'slider'

    def __init__(self, name, value, gui_element_type='slider', on_change=_empty_callback, increment=0.1, from_=0, to=10):
        self._increment = increment
        self.from_ = from_
        self.to = to
        super().__init__(name=name, value=value,
                         gui_element_type=gui_element_type, on_change=on_change)

    @property
    def increment(self):
        '''increment'''
        return self._increment

    @increment.setter
    def increment(self, inc):
        '''increment'''
        self._increment = float(inc)
        self.value = self.value

    @property
    def from_(self):
        '''min value'''
        return self._from

    @from_.setter
    def from_(self, from_):
        '''min value'''
        self._from = self.check(from_)

    @property
    def to(self):
        '''max value'''
        return self._to

    @to.setter
    def to(self, to):
        '''max value'''
        self._to = self.check(to)

    def round(self, val):
        '''Ensure that val is a multiple of "increment".'''
        if hasattr(self, '_increment'):
            val = round(float(val) / self.increment) * self.increment
        return val

    def check(self, val):
        '''Ensure that val is a float (convert it if necessary).'''
        return float(self.round(val))

    def convert_val(self, val):
        '''Ensure that val is a float and a multiple of "increment" (convert it if necessary).'''
        val = self.check(val)
        if val > self.to:
            val = self.to
        if val < self.from_:
            val = self.from_
        return val


class IntConfigItem(FloatConfigItem):
    def __init__(self, name, value, gui_element_type='slider', on_change=_empty_callback, increment=1, from_=0, to=100):
        FloatConfigItem.__init__(self, name=name, value=value,
                                 gui_element_type=gui_element_type, on_change=on_change,
                                 increment=increment, from_=from_, to=to)

    @property
    def increment(self):
        '''increment'''
        return self._increment

    @increment.setter
    def increment(self, inc):
        '''increment'''
        self._increment = int(inc)
        self.value = self.value

    def check(self, val):
        '''Ensure that val is an int (convert it if necessary).'''
        return int(self.round(val))
